"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os

from programy.config.base import BaseConfigurationData
from programy.storage.stores.file.engine import FileStorageEngine
from programy.storage.stores.file.store.config import FileStoreConfiguration
from programy.storage.stores.file.store.filestore import FileStore
from programy.utils.substitutions.substitues import Substitutions


class FileStorageConfiguration(BaseConfigurationData):

    @staticmethod
    def get_temp_dir():
        if os.name == 'posix':                                                  # pragma: no cover
            return '/tmp'                                                       # pragma: no cover
        elif os.name == 'nt':                                                   # pragma: no cover
            import tempfile                                                     # pragma: no cover
            return tempfile.gettempdir()                                        # pragma: no cover
        else:                                                                   # pragma: no cover
            raise Exception("Unknown operating system [%s]" % os.name)          # pragma: no cover

    def __init__(self, name="file"):
        BaseConfigurationData.__init__(self, name=name)

        tmpdir = FileStorageConfiguration.get_temp_dir()

        self._categories_storage = FileStoreConfiguration(dirs=[tmpdir + os.sep + "categories"], extension="aiml",
                                                          subdirs=True, fileformat="xml", encoding="utf-8",
                                                          delete_on_start=False)
        self._errors_storage = FileStoreConfiguration(file=tmpdir + os.sep + "debug/errors.txt", fileformat="text",
                                                      encoding="utf-8", delete_on_start=False)
        self._duplicates_storage = FileStoreConfiguration(file=tmpdir + os.sep + "debug/duplicates.txt",
                                                          fileformat="text", encoding="utf-8", delete_on_start=False)
        self._learnf_storage = FileStoreConfiguration(dirs=[tmpdir + os.sep + "categories/learnf"], extension="aiml",
                                                      subdirs=False, fileformat="xml", encoding="utf-8",
                                                      delete_on_start=False)

        self._conversation_storage = FileStoreConfiguration(dirs=[tmpdir + os.sep + "conversations"], extension="txt",
                                                            subdirs=False, fileformat="text", encoding="utf-8",
                                                            delete_on_start=False)

        self._sets_storage = FileStoreConfiguration(dirs=[tmpdir + os.sep + "sets"], extension="txt", subdirs=False,
                                                    fileformat="text", encoding="utf-8", delete_on_start=False)
        self._maps_storage = FileStoreConfiguration(dirs=[tmpdir + os.sep + "maps"], extension="txt", subdirs=False,
                                                    fileformat="text", encoding="utf-8", delete_on_start=False)
        self._rdf_storage = FileStoreConfiguration(dirs=[tmpdir + os.sep + "rdfs"], extension="txt", subdirs=True,
                                                   fileformat="text", encoding="utf-8", delete_on_start=False)

        self._denormal_storage = FileStoreConfiguration(file=tmpdir + os.sep + "lookups/denormal.txt",
                                                        fileformat="text", encoding="utf-8", delete_on_start=False)
        self._normal_storage = FileStoreConfiguration(file=tmpdir + os.sep + "lookups/normal.txt", fileformat="text",
                                                      encoding="utf-8", delete_on_start=False)
        self._gender_storage = FileStoreConfiguration(file=tmpdir + os.sep + "lookups/gender.txt", fileformat="text",
                                                      encoding="utf-8", delete_on_start=False)
        self._person_storage = FileStoreConfiguration(file=tmpdir + os.sep + "lookups/person.txt", fileformat="text",
                                                      encoding="utf-8", delete_on_start=False)
        self._person2_storage = FileStoreConfiguration(file=tmpdir + os.sep + "lookups/person2.txt", fileformat="text",
                                                       encoding="utf-8", delete_on_start=False)
        self._regex_storage = FileStoreConfiguration(file=tmpdir + os.sep + "regex/regex-templates.txt", fileformat="text",
                                                     encoding="utf-8", delete_on_start=False)

        self._properties_storage = FileStoreConfiguration(file=tmpdir + os.sep + "properties/properties.txt", fileformat="text",
                                                          encoding="utf-8", delete_on_start=False)
        self._defaults_storage = FileStoreConfiguration(file=tmpdir + os.sep + "properties/defaults.txt", fileformat="text",
                                                        encoding="utf-8", delete_on_start=False)

        self._twitter_storage = FileStoreConfiguration(dirs=[tmpdir + os.sep + "twitter"], extension="txt",
                                                       subdirs=False, fileformat="text", encoding="utf-8",
                                                       delete_on_start=False)

        self._spelling_storage = FileStoreConfiguration(file=tmpdir + os.sep + "spelling/corpus.txt", fileformat="text",
                                                        encoding="utf-8", delete_on_start=False)

        self._license_storage = FileStoreConfiguration(file=tmpdir + os.sep + "licenses/license.keys",
                                                       fileformat="text", encoding="utf-8", delete_on_start=False)

        self._pattern_nodes_storage = FileStoreConfiguration(file=tmpdir + os.sep + "nodes/pattern_nodes.conf",
                                                             fileformat="text", encoding="utf-8", delete_on_start=False)
        self._template_nodes_storage = FileStoreConfiguration(file=tmpdir + os.sep + "nodes/template_nodes.conf",
                                                              fileformat="text", encoding="utf-8",
                                                              delete_on_start=False)

        self._binaries_storage = FileStoreConfiguration(file=tmpdir + os.sep + "binaries/binaries.bin",
                                                        fileformat="binary", encoding="utf-8", delete_on_start=False)
        self._braintree_storage = FileStoreConfiguration(file=tmpdir + os.sep + "braintree/braintree.xml",
                                                         fileformat="xml", encoding="utf-8", delete_on_start=False)

        self._preprocessors_storage = FileStoreConfiguration(file=tmpdir + os.sep + "processing/preprocessors.conf",
                                                             fileformat="text", encoding="utf-8", delete_on_start=False)
        self._postprocessors_storage = FileStoreConfiguration(file=tmpdir + os.sep + "processing/postprocessors.conf",
                                                              fileformat="text", encoding="utf-8",
                                                              delete_on_start=False)
        self._postquestionprocessors_storage = FileStoreConfiguration(
            file=tmpdir + os.sep + "processing/postquestionprocessors.conf", fileformat="text", encoding="utf-8",
            delete_on_start=False)

        self._usergroups_storage = FileStoreConfiguration(file=tmpdir + os.sep + "security/usergroups.yaml",
                                                          fileformat="yaml", encoding="utf-8", delete_on_start=False)

        self._triggers_storage = FileStoreConfiguration(file=tmpdir + os.sep + "triggers/triggers.txt", fileformat="text",
                                                        encoding="utf-8", delete_on_start=False)

        self._oobs_storage = FileStoreConfiguration(file=tmpdir + os.sep + "oob/callmom.conf", fileformat="text",
                                                        encoding="utf-8", delete_on_start=False)

        self._services_storage = FileStoreConfiguration(dirs=[tmpdir + os.sep + "services"], extension="txt",
                                                          subdirs=False, fileformat="text", encoding="utf-8",
                                                          delete_on_start=False)

    @property
    def categories_storage(self):
        return self._categories_storage

    @property
    def errors_storage(self):
        return self._errors_storage

    @property
    def duplicates_storage(self):
        return self._duplicates_storage

    @property
    def learnf_storage(self):
        return self._learnf_storage

    @property
    def conversation_storage(self):
        return self._conversation_storage

    @property
    def sets_storage(self):
        return self._sets_storage

    @property
    def maps_storage(self):
        return self._maps_storage

    @property
    def rdf_storage(self):
        return self._rdf_storage

    @property
    def regex_storage(self):
        return self._regex_storage

    @property
    def denormal_storage(self):
        return self._denormal_storage

    @property
    def normal_storage(self):
        return self._normal_storage

    @property
    def gender_storage(self):
        return self._gender_storage

    @property
    def person_storage(self):
        return self._person_storage

    @property
    def person2_storage(self):
        return self._person2_storage

    @property
    def properties_storage(self):
        return self._properties_storage

    @property
    def defaults_storage(self):
        return self._defaults_storage

    @property
    def twitter_storage(self):
        return self._twitter_storage

    @property
    def spelling_storage(self):
        return self._spelling_storage

    @property
    def license_storage(self):
        return self._license_storage

    @property
    def pattern_nodes_storage(self):
        return self._pattern_nodes_storage

    @property
    def template_nodes_storage(self):
        return self._template_nodes_storage

    @property
    def binaries_storage(self):
        return self._binaries_storage

    @property
    def braintree_storage(self):
        return self._braintree_storage

    @property
    def preprocessors_storage(self):
        return self._preprocessors_storage

    @property
    def postprocessors_storage(self):
        return self._postprocessors_storage

    @property
    def postquestionprocessors_storage(self):
        return self._postquestionprocessors_storage

    @property
    def usergroups_storage(self):
        return self._usergroups_storage

    @property
    def triggers_storage(self):
        return self._triggers_storage

    @property
    def oobs_storage(self):
        return self._oobs_storage

    @property
    def services_storage(self):
        return self._services_storage

    def load_storage_config(self, storage_config, name, configuration_file, storage, bot_root,
                            subs: Substitutions = None):
        config_section = configuration_file.get_option(storage, 'config', subs=subs)
        storage_config_section = configuration_file.get_option(config_section, name, subs=subs)
        if storage_config_section is not None:
            storage_config.extract_configuration(configuration_file, storage_config_section, bot_root, subs=subs)

    def load_config_section(self, configuration_file, storage, bot_root, subs: Substitutions = None):
        if storage is not None:
            self.load_storage_config(self._categories_storage, FileStore.CATEGORIES_STORAGE, configuration_file,
                                     storage, bot_root, subs=subs)
            self.load_storage_config(self._errors_storage, FileStore.ERRORS_STORAGE, configuration_file, storage,
                                     bot_root, subs=subs)
            self.load_storage_config(self._duplicates_storage, FileStore.DUPLICATES_STORAGE, configuration_file,
                                     storage, bot_root, subs=subs)
            self.load_storage_config(self._learnf_storage, FileStore.LEARNF_STORAGE, configuration_file, storage,
                                     bot_root, subs=subs)

            self.load_storage_config(self._conversation_storage, FileStore.CONVERSATION_STORAGE, configuration_file,
                                     storage, bot_root, subs=subs)

            self.load_storage_config(self._sets_storage, FileStore.SETS_STORAGE, configuration_file, storage, bot_root,
                                     subs=subs)
            self.load_storage_config(self._maps_storage, FileStore.MAPS_STORAGE, configuration_file, storage, bot_root,
                                     subs=subs)
            self.load_storage_config(self._rdf_storage, FileStore.RDF_STORAGE, configuration_file, storage, bot_root,
                                     subs=subs)

            self.load_storage_config(self._denormal_storage, FileStore.DENORMAL_STORAGE, configuration_file, storage,
                                     bot_root, subs=subs)
            self.load_storage_config(self._normal_storage, FileStore.NORMAL_STORAGE, configuration_file, storage,
                                     bot_root, subs=subs)
            self.load_storage_config(self._gender_storage, FileStore.GENDER_STORAGE, configuration_file, storage,
                                     bot_root, subs=subs)
            self.load_storage_config(self._person_storage, FileStore.PERSON_STORAGE, configuration_file, storage,
                                     bot_root, subs=subs)
            self.load_storage_config(self._person2_storage, FileStore.PERSON2_STORAGE, configuration_file, storage,
                                     bot_root, subs=subs)
            self.load_storage_config(self._regex_storage, FileStore.REGEX_STORAGE, configuration_file, storage,
                                     bot_root, subs=subs)

            self.load_storage_config(self._properties_storage, FileStore.PROPERTIES_STORAGE, configuration_file,
                                     storage, bot_root, subs=subs)
            self.load_storage_config(self._defaults_storage, FileStore.DEFAULTS_STORAGE, configuration_file, storage,
                                     bot_root, subs=subs)

            self.load_storage_config(self._twitter_storage, FileStore.TWITTER_STORAGE, configuration_file, storage,
                                     bot_root, subs=subs)

            self.load_storage_config(self._spelling_storage, FileStore.SPELLING_STORAGE, configuration_file, storage,
                                     bot_root, subs=subs)

            self.load_storage_config(self._license_storage, FileStore.LICENSE_STORAGE, configuration_file, storage,
                                     bot_root, subs=subs)

            self.load_storage_config(self._pattern_nodes_storage, FileStore.PATTERN_NODES_STORAGE, configuration_file,
                                     storage, bot_root, subs=subs)
            self.load_storage_config(self._template_nodes_storage, FileStore.TEMPLATE_NODES_STORAGE, configuration_file,
                                     storage, bot_root, subs=subs)

            self.load_storage_config(self._binaries_storage, FileStore.BINARIES_STORAGE, configuration_file, storage,
                                     bot_root, subs=subs)
            self.load_storage_config(self._braintree_storage, FileStore.BRAINTREE_STORAGE, configuration_file, storage,
                                     bot_root, subs=subs)

            self.load_storage_config(self._preprocessors_storage, FileStore.PREPROCESSORS_STORAGE, configuration_file,
                                     storage, bot_root, subs=subs)
            self.load_storage_config(self._postprocessors_storage, FileStore.POSTPROCESSORS_STORAGE, configuration_file,
                                     storage, bot_root, subs=subs)
            self.load_storage_config(self._postquestionprocessors_storage, FileStore.POSTQUESTIONPROCESSORS_STORAGE,
                                     configuration_file, storage, bot_root, subs=subs)

            self.load_storage_config(self._usergroups_storage, FileStore.USERGROUPS_STORAGE, configuration_file,
                                     storage, bot_root, subs=subs)

            self.load_storage_config(self._triggers_storage, FileStore.TRIGGERS_STORAGE, configuration_file, storage,
                                     bot_root, subs=subs)

            self.load_storage_config(self._oobs_storage, FileStore.OOBS_STORAGE, configuration_file, storage,
                                     bot_root, subs=subs)

            self.load_storage_config(self._services_storage, FileStore.SERVICES_STORAGE, configuration_file, storage,
                                     bot_root, subs=subs)

    def create_filestorage_config(self):
        config = {}
        self._create_storage_map(config)
        return config

    def to_yaml(self, data, defaults=True):
        if defaults is True:
            self._create_storage_defaults(data)
        else:
            self._create_storage_map(data)

    def create_engine(self):
        engine = FileStorageEngine(self)
        engine.initialise()
        return engine

    def _create_storage_map(self, amap):

        amap[FileStore.CATEGORIES_STORAGE] = self._categories_storage
        amap[FileStore.ERRORS_STORAGE] = self._errors_storage
        amap[FileStore.DUPLICATES_STORAGE] = self._duplicates_storage
        amap[FileStore.LEARNF_STORAGE] = self._learnf_storage

        amap[FileStore.CONVERSATION_STORAGE] = self._conversation_storage

        amap[FileStore.SETS_STORAGE] = self._sets_storage
        amap[FileStore.MAPS_STORAGE] = self._maps_storage
        amap[FileStore.RDF_STORAGE] = self._rdf_storage

        amap[FileStore.DENORMAL_STORAGE] = self._denormal_storage
        amap[FileStore.NORMAL_STORAGE] = self._normal_storage
        amap[FileStore.GENDER_STORAGE] = self._gender_storage
        amap[FileStore.PERSON_STORAGE] = self._person_storage
        amap[FileStore.PERSON2_STORAGE] = self._person2_storage
        amap[FileStore.REGEX_STORAGE] = self._regex_storage

        amap[FileStore.PROPERTIES_STORAGE] = self._properties_storage
        amap[FileStore.DEFAULTS_STORAGE] = self._defaults_storage

        amap[FileStore.TWITTER_STORAGE] = self._twitter_storage

        amap[FileStore.SPELLING_STORAGE] = self._spelling_storage

        amap[FileStore.LICENSE_STORAGE] = self._license_storage

        amap[FileStore.PATTERN_NODES_STORAGE] = self._pattern_nodes_storage
        amap[FileStore.TEMPLATE_NODES_STORAGE] = self._template_nodes_storage

        amap[FileStore.BINARIES_STORAGE] = self._binaries_storage
        amap[FileStore.BRAINTREE_STORAGE] = self._braintree_storage

        amap[FileStore.PREPROCESSORS_STORAGE] = self._preprocessors_storage
        amap[FileStore.POSTPROCESSORS_STORAGE] = self._postprocessors_storage
        amap[FileStore.POSTQUESTIONPROCESSORS_STORAGE] = self._postquestionprocessors_storage

        amap[FileStore.USERGROUPS_STORAGE] = self._usergroups_storage

        amap[FileStore.TRIGGERS_STORAGE] = self._triggers_storage

        amap[FileStore.OOBS_STORAGE] = self._oobs_storage

        amap[FileStore.SERVICES_STORAGE] = self._services_storage

    def _create_storage_defaults(self, amap):

        tmpdir = FileStorageConfiguration.get_temp_dir()

        amap[FileStore.CATEGORIES_STORAGE] = {}
        config = FileStoreConfiguration(dirs=[tmpdir + os.sep + "categories"],
                                        extension="aiml", subdirs=True,
                                        fileformat="xml", encoding="utf-8",
                                        delete_on_start=False)
        config.to_yaml(amap[FileStore.CATEGORIES_STORAGE], defaults=False)

        amap[FileStore.ERRORS_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "debug/errors.txt",
                                        fileformat="text",
                                        encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.ERRORS_STORAGE], defaults=False)

        amap[FileStore.DUPLICATES_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "debug/duplicates.txt",
                                        fileformat="text",
                                        encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.DUPLICATES_STORAGE], defaults=False)

        amap[FileStore.LEARNF_STORAGE] = {}
        config = FileStoreConfiguration(dirs=[tmpdir + os.sep + "categories/learnf"],
                                        extension="aiml", subdirs=False,
                                        fileformat="xml", encoding="utf-8",
                                        delete_on_start=False)
        config.to_yaml(amap[FileStore.LEARNF_STORAGE], defaults=False)

        amap[FileStore.CONVERSATION_STORAGE] = {}
        config = FileStoreConfiguration(dirs=[tmpdir + os.sep + "conversations"],
                                        extension="txt",
                                        subdirs=False, fileformat="text",
                                        encoding="utf-8",
                                        delete_on_start=False)
        config.to_yaml(amap[FileStore.CONVERSATION_STORAGE], defaults=False)

        amap[FileStore.SETS_STORAGE] = {}
        config = FileStoreConfiguration(dirs=[tmpdir + os.sep + "sets"], extension="txt",
                                        subdirs=False,
                                        fileformat="text", encoding="utf-8",
                                        delete_on_start=False)
        config.to_yaml(amap[FileStore.SETS_STORAGE], defaults=False)

        amap[FileStore.MAPS_STORAGE] = {}
        config = FileStoreConfiguration(dirs=[tmpdir + os.sep + "maps"], extension="txt",
                                        subdirs=False,
                                        fileformat="text", encoding="utf-8",
                                        delete_on_start=False)
        config.to_yaml(amap[FileStore.MAPS_STORAGE], defaults=False)

        amap[FileStore.RDF_STORAGE] = {}
        config = FileStoreConfiguration(dirs=[tmpdir + os.sep + "rdfs"], extension="txt",
                                        subdirs=True,
                                        fileformat="text", encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.RDF_STORAGE], defaults=False)

        amap[FileStore.DENORMAL_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "lookups/denormal.txt",
                                        fileformat="text",
                                        encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.DENORMAL_STORAGE], defaults=False)

        amap[FileStore.NORMAL_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "lookups/normal.txt",
                                        fileformat="text",
                                        encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.NORMAL_STORAGE], defaults=False)

        amap[FileStore.GENDER_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "lookups/gender.txt",
                                        fileformat="text",
                                        encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.GENDER_STORAGE], defaults=False)

        amap[FileStore.PERSON_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "lookups/person.txt",
                                        fileformat="text",
                                        encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.PERSON_STORAGE], defaults=False)

        amap[FileStore.PERSON2_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "lookups/person2.txt",
                                        fileformat="text",
                                        encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.PERSON2_STORAGE], defaults=False)

        amap[FileStore.REGEX_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "regex/regex-templates.txt",
                                        fileformat="text",
                                        encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.REGEX_STORAGE], defaults=False)

        amap[FileStore.PROPERTIES_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "properties/properties.txt",
                                        fileformat="text",
                                        encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.PROPERTIES_STORAGE], defaults=False)

        amap[FileStore.DEFAULTS_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "properties/defaults.txt",
                                        fileformat="text",
                                        encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.DEFAULTS_STORAGE], defaults=False)

        amap[FileStore.TWITTER_STORAGE] = {}
        config = FileStoreConfiguration(dirs=[tmpdir + os.sep + "twitter"], extension="txt",
                                        subdirs=False,
                                        fileformat="text", encoding="utf-8",
                                        delete_on_start=False)
        config.to_yaml(amap[FileStore.TWITTER_STORAGE], defaults=False)

        amap[FileStore.SPELLING_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "spelling/corpus.txt",
                                                                  fileformat="text",
                                                                  encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.SPELLING_STORAGE], defaults=False)

        amap[FileStore.LICENSE_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "licenses/license.keys",
                                                                 fileformat="text",
                                                                 encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.LICENSE_STORAGE], defaults=False)

        amap[FileStore.PATTERN_NODES_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "nodes/pattern_nodes.conf",
                                                                       fileformat="text",
                                                                       encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.PATTERN_NODES_STORAGE], defaults=False)

        amap[FileStore.TEMPLATE_NODES_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "nodes/template_nodes.conf",
                                                                        fileformat="text",
                                                                        encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.TEMPLATE_NODES_STORAGE], defaults=False)

        amap[FileStore.BINARIES_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "braintree/braintree.bin",
                                       fileformat="binary",
                                       encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.BINARIES_STORAGE], defaults=False)

        amap[FileStore.BRAINTREE_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "braintree/braintree.xml",
                                        fileformat="xml",
                                        encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.BRAINTREE_STORAGE], defaults=False)

        amap[FileStore.PREPROCESSORS_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "processing/preprocessors.conf",
            fileformat="text", encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.PREPROCESSORS_STORAGE], defaults=False)

        amap[FileStore.POSTPROCESSORS_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "processing/postprocessors.conf",
            fileformat="text", encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.POSTPROCESSORS_STORAGE], defaults=False)

        amap[FileStore.POSTQUESTIONPROCESSORS_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "processing/postquestionprocessors.conf", fileformat="text",
            encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.POSTQUESTIONPROCESSORS_STORAGE], defaults=False)

        amap[FileStore.USERGROUPS_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "security/usergroups.yaml",
                                        fileformat="text",
                                        encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.USERGROUPS_STORAGE], defaults=False)

        amap[FileStore.TRIGGERS_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "triggers/triggers.txt",
                                        fileformat="text",
                                        encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.TRIGGERS_STORAGE], defaults=False)

        amap[FileStore.OOBS_STORAGE] = {}
        config = FileStoreConfiguration(file=tmpdir + os.sep + "oobs/callmom.conf",
                                        fileformat="text",
                                        encoding="utf-8", delete_on_start=False)
        config.to_yaml(amap[FileStore.OOBS_STORAGE], defaults=False)

        amap[FileStore.SERVICES_STORAGE] = {}
        config = FileStoreConfiguration(dirs=[tmpdir + os.sep + "services"],
                                        extension="txt", subdirs=True,
                                        fileformat="text", encoding="utf-8",
                                        delete_on_start=False)
        config.to_yaml(amap[FileStore.SERVICES_STORAGE], defaults=False)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation dirs (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.config.section import BaseSectionConfigurationData
from programy.utils.substitutions.substitues import Substitutions


class FileStoreConfiguration(BaseSectionConfigurationData):

    def __init__(self, name='storage', dirs=None, extension=None, subdirs=False, fileformat=None, encoding=None,
                 delete_on_start=False, file=None):
        BaseSectionConfigurationData.__init__(self, name)

        self._dirs = None
        self._has_single_file = False

        if dirs is not None:
            if isinstance(dirs, (list,)) is False:
                self._dirs = [dirs]
                self._has_single_file = True

            else:
                self._dirs = dirs[:]
                self._has_single_file = False

        elif file is not None:
            self._dirs = [file]
            self._has_single_file = True

        self._extension = extension
        self._subdirs = subdirs

        self._format = fileformat
        self._encoding = encoding

        self._delete_on_start = delete_on_start

    def has_multiple_dirs(self):
        if self._has_single_file:
            return False
        else:
            return True

    def has_single_file(self):
        return self._has_single_file

    @property
    def file(self):
        return self._dirs[0]

    @property
    def dirs(self):
        return self._dirs

    @property
    def extension(self):
        return self._extension

    @property
    def subdirs(self):
        return self._subdirs

    @property
    def format(self):
        return self._format

    @property
    def encoding(self):
        return self._encoding

    @property
    def delete_on_start(self):
        return self._delete_on_start

    def load_config_section(self, configuration_file, configuration, bot_root, subs: Substitutions = None):
        dirs_config = configuration_file.get_option(configuration, self.section_name)
        if dirs_config is not None:
            self.extract_configuration(configuration_file, dirs_config, bot_root, subs=subs)
        else:
            YLogger.warning(self, "'%s' section missing from bot config, using to defaults", self.section_name)

    def extract_configuration(self, configuration_file, files_config, bot_root, subs: Substitutions = None):
        dirs = configuration_file.get_multi_file_option(files_config, "dirs", bot_root, subs=subs)
        if dirs is not None and dirs:
            self._dirs = dirs
            self._extension = configuration_file.get_option(files_config, "extension", subs=subs)
            self._subdirs = configuration_file.get_bool_option(files_config, "subdirs", False, subs=subs)
            self._has_single_file = False
        else:
            file = configuration_file.get_option(files_config, "file", subs=subs)
            if file is not None:
                self._dirs = [self.sub_bot_root(file, bot_root)]
                self._has_single_file = True

        self._format = configuration_file.get_option(files_config, "format", None, subs=subs)
        self._encoding = configuration_file.get_option(files_config, "encoding", None, subs=subs)

        self._delete_on_start = configuration_file.get_bool_option(files_config, "delete_on_start", False, subs=subs)

    def to_yaml(self, data, defaults=True):
        if defaults is True:

            if self._has_single_file is True:
                data['file'] = "./storage/%s" % self.id

            else:
                data['dirs'] = "./storage/%s" % self.id

            data['extension'] = ".txt"
            data['subdirs'] = False

            data['format'] = None
            data['encoding'] = None
            data['delete_on_start'] = False

        else:
            if self._has_single_file is True:
                data['file'] = self._dirs[0]

            else:
                data['dirs'] = self._dirs

            data['extension'] = self._extension
            data['subdirs'] = self._subdirs

            data['format'] = self._format
            data['encoding'] = self._encoding
            data['delete_on_start'] = self._delete_on_start



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from programy.utils.logging.ylogger import YLogger

from programy.utils.classes.loader import ClassLoader
from programy.dynamic.sets.numeric import IsNumeric
from programy.dynamic.maps.plural import PluralMap
from programy.dynamic.maps.singular import SingularMap
from programy.dynamic.maps.predecessor import PredecessorMap
from programy.dynamic.maps.successor import SuccessorMap


class DynamicsCollection:

    def __init__(self):
        self._dynamic_sets = {}
        self._dynamic_maps = {}
        self._dynamic_vars = {}

    def load_from_configuration(self, dynamics_configuration):

        if dynamics_configuration is not None:

            for set_name in dynamics_configuration.dynamic_sets:
                self.add_dynamic_set(set_name, dynamics_configuration.dynamic_sets[set_name], dynamics_configuration)

            for map_name in dynamics_configuration.dynamic_maps:
                self.add_dynamic_map(map_name, dynamics_configuration.dynamic_maps[map_name], dynamics_configuration)

            for var_name in dynamics_configuration.dynamic_vars:
                self.add_dynamic_var(var_name, dynamics_configuration.dynamic_vars[var_name], dynamics_configuration)

            self.load_default_dynamics(dynamics_configuration)

    def load_default_dynamics(self, dynamics_configuration):
        self.load_default_dynamic_sets(dynamics_configuration)
        self.load_default_dynamic_maps(dynamics_configuration)
        self.load_default_dynamic_vars(dynamics_configuration)

    def load_default_dynamic_sets(self, dynamics_configuration):
        if IsNumeric.NAME not in self._dynamic_sets:
            YLogger.warning(self, "Dynamic set %s not defined, adding default implementation", IsNumeric.NAME)
            self._dynamic_sets[IsNumeric.NAME] = IsNumeric(dynamics_configuration)

    def load_default_dynamic_maps(self, dynamics_configuration):
        if PluralMap.NAME not in self._dynamic_maps:
            YLogger.warning(self, "Dynamic set %s not defined, adding default implementation", PluralMap.NAME)
            self._dynamic_maps[PluralMap.NAME] = PluralMap(dynamics_configuration)

        if SingularMap.NAME not in self._dynamic_maps:
            YLogger.warning(self, "Dynamic set %s not defined, adding default implementation", SingularMap.NAME)
            self._dynamic_maps[SingularMap.NAME] = SingularMap(dynamics_configuration)

        if SuccessorMap.NAME not in self._dynamic_maps:
            YLogger.warning(self, "Dynamic set %s not defined, adding default implementation", SuccessorMap.NAME)
            self._dynamic_maps[SuccessorMap.NAME] = SuccessorMap(dynamics_configuration)

        if PredecessorMap.NAME not in self._dynamic_maps:
            YLogger.warning(self, "Dynamic set %s not defined, adding default implementation", PredecessorMap.NAME)
            self._dynamic_maps[PredecessorMap.NAME] = PredecessorMap(dynamics_configuration)

    def load_default_dynamic_vars(self, dynamics_configuration):
        del dynamics_configuration
        return

    ###################################################################################################
    # Dynamic Sets

    @property
    def dynamic_sets(self):
        return self._dynamic_sets

    def add_dynamic_set(self, name, classname, config_file):
        self._dynamic_sets[name.upper()] = (ClassLoader.instantiate_class(classname))(config_file)

    def is_dynamic_set(self, name):
        return bool(name.upper() in self._dynamic_sets)

    def dynamic_set(self, client_context, name, value, additional=None):
        name = name.upper()
        if name in self._dynamic_sets:
            dynamic_set = self._dynamic_sets[name]
            return dynamic_set.is_member(client_context, value, additional)
        return None

    ###################################################################################################
    # Dynamic Maps

    @property
    def dynamic_maps(self):
        return self._dynamic_maps

    def add_dynamic_map(self, name, classname, config_file):
        self._dynamic_maps[name.upper()] = (ClassLoader.instantiate_class(classname))(config_file)

    def is_dynamic_map(self, name):
        return bool(name.upper() in self._dynamic_maps)

    def dynamic_map(self, client_context, name, value):
        name = name.upper()
        if name in self._dynamic_maps:
            dynamic_map = self._dynamic_maps[name]
            return dynamic_map.map_value(client_context, value)
        return None

    ###################################################################################################
    # Dynamic Vars

    @property
    def dynamic_vars(self):
        return self._dynamic_vars

    def add_dynamic_var(self, name, classname, config_file):
        self._dynamic_vars[name.upper()] = (ClassLoader.instantiate_class(classname))(config_file)

    def is_dynamic_var(self, name):
        return bool(name.upper() in self._dynamic_vars)

    def dynamic_var(self, client_context, name, value=None):
        name = name.upper()
        if name in self._dynamic_vars:
            dynamic_var = self._dynamic_vars[name]
            return dynamic_var.get_value(client_context, value)

        return None

    def _set_value(self, client_context, name, value):
        name = name.upper()
        if name in self._dynamic_vars:
            dynamic_var = self._dynamic_vars[name]
            dynamic_var.set_value(client_context, value)
            return True

        return False

    def set_dynamic_var(self, client_context, name, value):
        try:
            return self._set_value(client_context, name, value)

        except Exception as e:
            YLogger.exception(self, "Unable to set value for dynamic var %s", e, name)

        return False



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import json
from programy.utils.logging.ylogger import YLogger
from programy.config.file.file import BaseConfigurationFile
from programy.config.programy import ProgramyConfiguration
from programy.utils.substitutions.substitues import Substitutions


class JSONConfigurationFile(BaseConfigurationFile):

    def __init__(self):
        BaseConfigurationFile.__init__(self)
        self.json_data = None

    def load_from_text(self, text, client_configuration, bot_root, subs: Substitutions = None):
        self.json_data = json.loads(text)
        configuration = ProgramyConfiguration(client_configuration)
        configuration.load_config_data(self, bot_root, subs)
        return configuration

    def load_from_file(self, filename, client_configuration, bot_root, subs: Substitutions = None):
        configuration = ProgramyConfiguration(client_configuration)
        try:
            with open(filename, 'r+', encoding="utf-8") as json_data_file:
                self.json_data = json.load(json_data_file)
                configuration.load_config_data(self, bot_root, subs)

        except Exception as excep:
            YLogger.exception(self, "Failed to open json config file [%s]", excep, filename)

        return configuration

    def get_section(self, section_name, parent_section=None):
        if parent_section is None:
            if section_name in self.json_data:
                return self.json_data[section_name]
        else:
            if section_name in parent_section:
                return parent_section[section_name]
        return None

    def get_keys(self, section):
        return section.keys()

    def get_child_section_keys(self, child_section_name, parent_section):
        if child_section_name in parent_section:
            return parent_section[child_section_name].keys()

        return None

    def get_option(self, section, option_name, missing_value=None, subs: Substitutions = None):
        if option_name in section:
            option_value = section[option_name]
            return self._replace_subs(subs, option_value)

        YLogger.warning(self, "Missing value for [%s] in config , return default value %s", option_name, missing_value)
        return missing_value

    def get_bool_option(self, section, option_name, missing_value=False, subs: Substitutions = None):
        if option_name in section:
            option_value = section[option_name]
            if isinstance(option_value, bool):
                return option_value

            return bool(self._replace_subs(subs, option_value))

        YLogger.warning(self, "Missing value for [%s] in config, return default value %s", option_name, missing_value)
        return missing_value

    def get_int_option(self, section, option_name, missing_value=0, subs: Substitutions = None):
        if option_name in section:
            option_value = section[option_name]
            if isinstance(option_value, int):
                return int(option_value)

        if missing_value is not None:
            YLogger.warning(self, "Missing value for [%s] in config, return default value %d", option_name,
                            missing_value)
        else:
            YLogger.warning(self, "Missing value for [%s] in config, return default value None", option_name)

        return missing_value

    def get_multi_option(self, section, option_name, missing_value=[], subs: Substitutions = None):

        value = self.get_option(section, option_name, missing_value)
        if isinstance(value, list):
            values = value

        else:
            values = [value]

        multis = []
        for value in values:
            multis.append(self._replace_subs(subs, value))

        return multis

    def get_multi_file_option(self, section, option_name, bot_root, missing_value=[], subs: Substitutions = None):

        value = self.get_option(section, option_name, missing_value)
        if isinstance(value, list):
            values = value

        else:
            values = [value]

        multis = []
        for value in values:
            if isinstance(value, str):
                value = self._replace_subs(subs, value)
                multis.append(value.replace('$BOT_ROOT', bot_root))

        return multis



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.parser.template.nodes.base import TemplateNode
from programy.services.service import ServiceFactory
from programy.parser.exceptions import ParserException
from programy.utils.text.text import TextUtils


class TemplateSRAIXNode(TemplateNode):

    def __init__(self):
        TemplateNode.__init__(self)
        self._service = None

    @property
    def service(self):
        return self._service

    @service.setter
    def service(self, service):
        self._service = service

    def resolve_to_string(self, client_context):
        resolved = self.resolve_children_to_string(client_context)
        YLogger.debug(client_context, "[%s] resolved to [%s]", self.to_string(), resolved)

        if self._service is not None:
            bot_service = ServiceFactory.get_service(self._service)
            response = bot_service.ask_question(client_context, resolved)
            YLogger.debug(client_context, "SRAIX service [%s] return [%s]", self._service, response)
            return response
        else:
            YLogger.error(client_context, "Sorry SRAIX does not currently have an implementation for [%s]",
                          self._service)
            return ""

    def to_string(self):
        if self._service is not None:
            return "[SRAIX (service=%s)]" % (self._service)
        return "[SRAIX ()]"

    def to_xml(self, client_context):
        xml = '<sraix'
        if self._service is not None:
            xml += ' service="%s"' % self.service
        xml += '>'
        xml += self.children_to_xml(client_context)
        xml += '</sraix>'
        return xml

    #######################################################################################################
    # SRAIX_ATTRIBUTES ::= host="HOSTNAME" | botid="BOTID" | hint="TEXT" | apikey="APIKEY" | service="SERVICE"
    # SRAIX_ATTRIBUTE_TAGS ::= <host>TEMPLATE_EXPRESSION</host> | <botid>TEMPLATE_EXPRESSION</botid> |
    # <hint>TEMPLATE_EXPRESSION</hint> | <apikey>TEMPLATE_EXPRESSION</apikey> | <service>TEMPLATE_EXPRESSION</service>
    # SRAIX_EXPRESSION ::== <sraix( SRAIX_ATTRIBUTES)*>TEMPLATE_EXPRESSION</sraix> |

    def parse_expression(self, graph, expression):

        if 'host' in expression.attrib:
            YLogger.warning(self, "'host' attrib not supported in sraix, moved to config, see documentation")
        if 'botid' in expression.attrib:
            YLogger.warning(self, "'botid' attrib not supported in sraix, moved to config, see documentation")
        if 'hint' in expression.attrib:
            YLogger.warning(self, "'hint' attrib not supported in sraix, moved to config, see documentation")
        if 'apikey' in expression.attrib:
            YLogger.warning(self, "'apikey' attrib not supported in sraix, moved to config, see documentation")

        if 'service' in expression.attrib:
            self.service = expression.attrib['service']

        head_text = self.get_text_from_element(expression)
        self.parse_text(graph, head_text)

        for child in expression:
            tag_name = TextUtils.tag_from_text(child.tag)

            if tag_name == 'host':
                YLogger.warning(self, "'host' element not supported in sraix, moved to config, see documentation")
            elif tag_name == 'botid':
                YLogger.warning(self, "'botid' element not supported in sraix, moved to config, see documentation")
            elif tag_name == 'hint':
                YLogger.warning(self, "'hint' element not supported in sraix, moved to config, see documentation")
            elif tag_name == 'apikey':
                YLogger.warning(self, "'apikey' element not supported in sraix, moved to config, see documentation")
            elif tag_name == 'service':
                self.service = self.get_text_from_element(child)
            else:
                graph.parse_tag_expression(child, self)

            tail_text = self.get_tail_from_element(child)
            self.parse_text(graph, tail_text)

        if self.service is None:
            raise ParserException("SRAIX node, service attribute missing !")



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import sleekxmpp
from programy.utils.logging.ylogger import YLogger


class XmppClient(sleekxmpp.ClientXMPP):

    def __init__(self, bot_client, jid, password):
        self._bot_client = bot_client
        sleekxmpp.ClientXMPP.__init__(self, jid, password)
        self.add_event_handlers()

    def add_event_handlers(self):
        YLogger.debug(self, "XMPPClient adding event handlers")
        self.add_event_handler("session_start", self.start)
        self.add_event_handler("message", self.message)

    def register_xep_plugins(self, configuration):
        YLogger.debug(self, "XMPPClient registering XEP plugins")

        if configuration.xep_0030 is True:
            YLogger.debug(self, "XMPPClient registering xep_0030 plugin")
            self.register_plugin('xep_0030')

        if configuration.xep_0004 is True:
            YLogger.debug(self, "XMPPClient registering xep_0004 plugin")
            self.register_plugin('xep_0004')

        if configuration.xep_0060 is True:
            YLogger.debug(self, "XMPPClient registering xep_0060 plugin")
            self.register_plugin('xep_0060')

        if configuration.xep_0199 is True:
            YLogger.debug(self, "XMPPClient registering xep_0199 plugin")
            self.register_plugin('xep_0199')


    def start(self, event):
        YLogger.debug(self, "XMPPClient starting....[%s]", str(event))
        self.send_presence()
        self.get_roster()

    def is_valid_message(self, msg):
        return bool(msg['type'] in ('chat', 'normal'))

    def get_question(self, msg):
        return msg['body']

    def get_userid(self, msg):
        return msg['from'].bare

    def send_response(self, msg, response):
        if response is not None:
            msg.reply(response).send()

    def message(self, msg):
        print("Msg = ", msg)
        if self.is_valid_message(msg) is True:

            question = self.get_question(msg)
            if question is None:
                YLogger.debug(self, "XMPPClient - Missing 'question' from XMPP message")
                return

            userid = self.get_userid(msg)
            if userid is None:
                YLogger.debug(self, "XMPCLient - Missing 'userid' from XMPP message")
                return

            response = self._bot_client.ask_question(userid, question)

            self.send_response(msg, response)

        else:
            YLogger.debug(self, "Invalid XMPP message")
            self.send_response(msg, "Sorry, no idea!")



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.clients.render.text import TextRenderer


class HtmlRenderer(TextRenderer):

    def __init__(self, callback=None):
        TextRenderer.__init__(self, callback)

    def create_postback_url(self):
        return "#"

    def handle_text(self, client_context, text, response=True):
        rendered = "%s" % text['text']
        if self._client and response is True:
            self._client.process_response(client_context, rendered)
        return rendered

    def _add_class_and_id(self, data, rendered):
        toreturn = ""

        if 'class' in data:
            toreturn += ' class="%s programy"'%data['class']
        else:
            toreturn += ' class="programy"'

        if 'id' in data:
            toreturn += ' id="%s"'%data['id']

        return toreturn

    def handle_url_button(self, client_context, button, response=True):
        rendered = '<a'
        rendered += self._add_class_and_id(button, rendered)
        rendered += ' href="%s">%s</a>' % (button['url'], button['text'])
        if self._client and response is True:
            self._client.process_response(client_context, rendered)
        return rendered

    def handle_postback_button(self, client_context, button, response=True):
        rendered = '<a'
        rendered += self._add_class_and_id(button, rendered)
        rendered += ' postback="%s" href="#">%s</a>' % (button['postback'], button['text'])
        if self._client and response is True:
            self._client.process_response(client_context, rendered)
        return rendered

    def handle_link(self, client_context, link, response=True):
        rendered = '<a'
        rendered += self._add_class_and_id(link, rendered)
        rendered += ' href="%s">%s</a>' % (link['url'], link['text'])
        if self._client and response is True:
            self._client.process_response(client_context, rendered)
        return rendered

    def handle_image(self, client_context, image, response=True):
        rendered = '<img'
        rendered += self._add_class_and_id(image, rendered)
        rendered += ' src="%s" />' % image['url']
        if self._client and response is True:
            self._client.process_response(client_context, rendered)
        return rendered

    def handle_video(self, client_context, video, response=True):
        rendered = '<video'
        rendered += self._add_class_and_id(video, rendered)
        rendered += ' src="%s">'%video['url']
        rendered += """\nSorry, your browser doesn't support embedded videos, 
but don't worry, you can <a href="%s">download it</a>
and watch it with your favorite video player!
</video>""" %video['url']
        if self._client and response is True:
            self._client.process_response(client_context, rendered)
        return rendered

    def _format_card(self, client_context, card):
        rendered = '<div'
        rendered += self._add_class_and_id(card, rendered)
        rendered += '>'
        rendered += '<img src="%s" />' % card['image']
        rendered += '<h1>%s</h1>' % card['title']
        rendered += '<h2>%s</h2>' % card['subtitle']
        for button in card['buttons']:
            if button['url'] is not None:
                rendered += '<a href="%s">%s</a>' % (button['url'], button['text'])
            else:
                rendered += '<a postback="%s" href="#">%s</a>' % (button['postback'], button['text'])
        rendered += '</div>'
        return rendered

    def handle_card(self, client_context, card, response=True):
        rendered = self._format_card(client_context, card)
        if self._client and response is True:
            self._client.process_response(client_context, rendered)
        return rendered

    def handle_carousel(self, client_context, carousel, response=True):
        rendered = '<div'
        rendered += self._add_class_and_id(carousel, rendered)
        rendered += '>'
        for card in carousel['cards']:
            rendered += self._format_card(client_context, card)
        rendered += "</div>"
        if self._client and response is True:
            self._client.process_response(client_context, rendered)
        return rendered

    def handle_reply(self, client_context, reply, response=True):
        rendered = ''
        if reply['postback'] is not None:
            rendered += '<a'
            rendered += self._add_class_and_id(reply, rendered)
            rendered += ' postback="%s" href="#">%s</a>' % (reply['postback'], reply['text'])
        else:
            rendered += '<a'
            rendered += self._add_class_and_id(reply, rendered)
            rendered += ' postback="%s" href="#">%s</a>' % (reply['text'], reply['text'])
        if self._client and response is True:
            self._client.process_response(client_context, rendered)
        return rendered

    def handle_delay(self, client_context, delay, response=True):
        rendered = '<div'
        rendered += self._add_class_and_id(delay, rendered)
        rendered += '>...</div>'
        if self._client and response is True:
            self._client.process_response(client_context, rendered)
        return rendered

    def handle_split(self, client_context, split, response=True):
        rendered = "<br"
        rendered += self._add_class_and_id(split, rendered)
        rendered += " />"
        if self._client and response is True:
            self._client.process_response(client_context, rendered)
        return rendered

    def handle_list(self, client_context, lst, response=True):
        rendered = "<ul"
        rendered += self._add_class_and_id(lst, rendered)
        rendered += ">"
        for item in lst.get('items', []):
            rendered += "<li>%s</li>" % self._handle_children(client_context, item)
        rendered += "</ul>"
        if self._client and response is True:
            self._client.process_response(client_context, rendered)
        return rendered

    def handle_ordered_list(self, client_context, lst, response=True):
        rendered = "<ol"
        rendered += self._add_class_and_id(lst, rendered)
        rendered += ">"
        for item in lst.get('items', []):
            rendered += "<li>%s</li>" % self._handle_children(client_context, item)
        rendered += "</ol>"
        if self._client and response is True:
            self._client.process_response(client_context, rendered)
        return rendered

    def handle_location(self, client_context, location, response=True):
        rendered = ""
        if self._client and response is True:
            self._client.process_response(client_context, rendered)
        return rendered

    def handle_xml(self, client_context, item, response=True):
        rendered = item['text']
        if self._client and response is True:
            self._client.process_response(client_context, rendered)
        return rendered

    def _handle_children(self, client_context, item):

        if item['type'] == 'button':
            return self.handle_button(client_context, item, response=False)

        elif item['type'] == 'link':
            return self.handle_link(client_context, item, response=False)

        elif item['type'] == 'image':
            return self.handle_image(client_context, item, response=False)

        elif item['type'] == 'video':
            return self.handle_video(client_context, item, response=False)

        elif item['type'] == 'card':
            return self.handle_card(client_context, item, response=False)

        elif item['type'] == 'carousel':
            return self.handle_carousel(client_context, item, response=False)

        elif item['type'] == 'reply':
            return self.handle_reply(client_context, item, response=False)

        elif item['type'] == 'delay':
            return self.handle_delay(client_context, item, response=False)

        elif item['type'] == 'split':
            return self.handle_split(client_context, item, response=False)

        elif item['type'] == 'list':
            return self.handle_list(client_context, item, response=False)

        elif item['type'] == 'olist':
            return self.handle_ordered_list(client_context, item, response=False)

        elif item['type'] == 'location':
            return self.handle_location(client_context, item, response=False)

        elif item['type'] == 'tts':
            return self.handle_tts(client_context, item, response=False)

        elif item['type'] == 'text':
            return self.handle_text(client_context, item, response=False)

        return self.handle_xml(client_context, item, response=False)




"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import requests
import json
"""
import json
import requests
from flask import Flask, request
from flask import current_app

servers = {}

if __name__ == '__main__':

    print("Initiating Healthcheck App...")

    APP = Flask(__name__)

    @APP.route('/')
    def index():
        return current_app.send_static_file('healthcheck.html')

    # name=NAME, host=HOST, port=PORT
    @APP.route('/api/health/v1.0/register')
    def regiter():
        name = request.args.get('name')
        host = request.args.get('host')
        port = request.args.get('port')
        url = request.args.get('url')

        if name in servers:
            print("Re-register existing server [%s]" % name)
        else:
            print("Register server [%s]" % name)

        if url.startswith('/') is False:
            url = "/" + url

        servers[name] = "http://%s:%s%s"% (host, port, url)
        print(servers[name])

        return "OK"

    # name=NAME
    @APP.route('/api/health/v1.0/unregister')
    def unregiter():
        name = request.args.get('name')
        if name in servers:
            del servers[name]
            print("Unregistered server [%s]"%name)
        else:
            print("Unable to register server, unknown name [%s]" % name)
        return "OK"

    @APP.route('/api/health/v1.0/ping')
    def ping():
        healthchecks = get_health_check_from_servers(servers)
        return healthchecks_to_treedata(healthchecks)

    def healthchecks_to_treedata(healthchecks):
        treedata = {"source": []}


        if 'pings' in healthchecks:
            healthcheck_pings = healthchecks['pings']

            id = 1
            for healthcheck in healthcheck_pings :

                client = {"title": healthcheck["name"], "key": str(id)}
                id += 1

                bots = {"title": "Bots", "folder": True, "key": str(id), "children": []}
                id += 1

                logging = {"title": "Logging", "folder": True, "key": str(id), "children": []}
                id += 1

                if 'ping' in healthcheck:
                    ping = healthcheck['ping']

                    if 'logging' in ping:
                        logging["children"].append({"title": "criticals: " + str(ping['logging']['criticals']), "key": str(id)})
                        id += 1
                        logging["children"].append({"title": "exceptions: " + str(ping['logging']['exceptions']), "key": str(id)})
                        id += 1
                        logging["children"].append({"title": "fatals: " + str(ping['logging']['fatals']), "key": str(id)})
                        id += 1
                        logging["children"].append({"title": "errors: " + str(ping['logging']['errors']), "key": str(id)})
                        id += 1
                        logging["children"].append({"title": "warnings: " + str(ping['logging']['warnings']), "key": str(id)})
                        id += 1
                        logging["children"].append({"title": "debugs: " + str(ping['logging']['debugs']), "key": str(id)})
                        id += 1
                        logging["children"].append({"title": "infos: " + str(ping['logging']['infos']), "key": str(id)})
                        id += 1

                    type = {"title": "Type: " + ping['client'], "key": str(id)}
                    id += 1

                    questions = {"title": "Questions: " + str(ping['questions']), "key": str(id)}
                    id += 1

                    starttime = {"title": "Start Time: " + ping['start_time'], "key": str(id)}
                    id += 1

                    if 'bots' in ping:
                        for bot in ping['bots']:
                            abot = {"title": bot['id'], "folder": True, "key": str(id), "children": []}
                            id += 1

                            abot['children'].append({"title": "Questions: " + str(bot['questions']), "key": str(id)})
                            id += 1

                            brains = {"title": "Brains", "folder": True, "key": str(id), "children": []}
                            id += 1

                            bots['children'].append(abot)
                            abot['children'].append(brains)

                            if 'brains' in bot:
                                for brain in bot['brains']:
                                    abrain = {"title": brain['id'], "folder": True, "key": str(id), "children": []}
                                    id += 1
                                    brains['children'].append(abrain)
                                    abrain['children'].append({"title": "Questions: " + str(brain['questions']), "key": str(id)})
                                    id += 1

                    client['children'] = [type, questions, starttime, bots, logging]

                treedata['source'].append(client)

        return json.dumps(treedata)


    # noinspection Pylint
    def get_health_check_from_serversX(servers):
        del servers

        data = {'pings': [{'name': 'webchat1', 'ping': {'bots': [{'brains': [{'id': 'brain', 'questions': 0}], 'id': 'bot', 'questions': 0}], 'client': 'WebChat', 'logging': {'criticals': 0, 'debugs': 8249, 'errors': 50, 'exceptions': 1, 'fatals': 0, 'infos': 842, 'warnings': 446}, 'questions': 0, 'start_time': '2019-03-18 19:55:38.023020'}},
                          {'name': '2brains', 'ping': {
                              'bots': [{'brains': [{'id': 'brain1', 'questions': 0}, {'id': 'brain2', 'questions': 0}], 'id': 'bot', 'questions': 0}],
                              'client': 'WebChat',
                              'logging': {'criticals': 0, 'debugs': 8249, 'errors': 50, 'exceptions': 1, 'fatals': 0,
                                          'infos': 842, 'warnings': 446}, 'questions': 0,
                              'start_time': '2019-03-18 19:55:38.023020'}},
                          {'name': '2bots', 'ping': {
                              'bots': [{'brains': [{'id': 'brain', 'questions': 0}], 'id': 'bot1', 'questions': 0},
                                       {'brains': [{'id': 'brain', 'questions': 0}], 'id': 'bot2', 'questions': 0}],
                              'client': 'WebChat',
                              'logging': {'criticals': 0, 'debugs': 8249, 'errors': 50, 'exceptions': 1, 'fatals': 0,
                                          'infos': 842, 'warnings': 446}, 'questions': 0,
                              'start_time': '2019-03-18 19:55:38.023020'}},
                          {'name': '2bots2brains', 'ping': {
                              'bots': [{'brains': [{'id': 'brain1', 'questions': 0}, {'id': 'brain2', 'questions': 0}], 'id': 'bot1', 'questions': 0},
                                       {'brains': [{'id': 'brain1', 'questions': 0}, {'id': 'brain2', 'questions': 0}], 'id': 'bot2', 'questions': 0}],
                              'client': 'WebChat',
                              'logging': {'criticals': 0, 'debugs': 8249, 'errors': 50, 'exceptions': 1, 'fatals': 0,
                                          'infos': 842, 'warnings': 446}, 'questions': 0,
                              'start_time': '2019-03-18 19:55:38.023020'}},

                          ]}

        return data


    def get_health_check_from_servers(servers):
        result = {"pings": []}

        for name, url in servers.items():
            healthcheck = {"name": name}
            ping = requests.get(url)
            healthcheck['ping'] = ping.json()
            result['pings'].append(healthcheck)

        return result

    APP.run()



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.parser.exceptions import ParserException
from programy.utils.text.text import TextUtils


class TemplateGraph:

    def __init__(self, aiml_parser):
        self._aiml_parser = aiml_parser
        self._template_factory = aiml_parser.brain.template_factory

    @property
    def aiml_parser(self):
        return self._aiml_parser

    @property
    def template_factory(self):
        return self._template_factory

    #
    # TEMPLATE_EXPRESSION ::== TEXT | TAG_EXPRESSION | (TEMPLATE_EXPRESSION)*
    #
    def parse_template_expression(self, pattern):
        node = self.get_base_node()
        node.parse_template_node(self, pattern)
        return node

    def get_node_class_by_name(self, name):
        if self._template_factory.exists(name):
            return self._template_factory.new_node_class(name)
        else:
            raise ParserException("No node [%s] registered in Template Node Factory" % (name))

    # Helper function to return TemplateNode
    def get_base_node(self):
        base_class = self.get_node_class_by_name('base')
        return base_class()

    # Helper function to return TemplateWordNode
    def get_word_node(self, text):
        word_class = self.get_node_class_by_name('word')
        return word_class(text)

    def parse_tag_expression(self, expression, branch):
        tag_name = TextUtils.tag_from_text(expression.tag)
        if self._template_factory.exists(tag_name):
            if tag_name == "condition":
                node_instance = self._template_factory.new_node_class(tag_name)()
            else:
                node_instance = self._template_factory.new_node_class(tag_name)()
            node_instance.parse_expression(self, expression)
            branch.children.append(node_instance)
        else:
            self.parse_unknown_as_xml_node(expression, branch)

    #######################################################################################################
    # 	UNKNONWN NODE
    #   When its a node we don't know, add it as a text node. This deals with html nodes creeping into the text
    def parse_unknown_as_xml_node(self, expression, branch):
        xml_node_class = self.get_node_class_by_name('xml')
        xml_node = xml_node_class()
        branch.children.append(xml_node)
        xml_node.parse_expression(self, expression)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger

from programy.config.base import BaseConfigurationData
from programy.utils.substitutions.substitues import Substitutions


class SchedulerJobStoreConfiguration(BaseConfigurationData):

    def __init__(self):
        BaseConfigurationData.__init__(self, name="jobstore")
        self._name = None
        self._storage = None

    @property
    def name(self):
        return self._name

    @property
    def storage(self):
        return self._storage

    def load_config_section(self, configuration_file, configuration, bot_root, subs: Substitutions = None):
        del subs
        jobstore = configuration_file.get_section(self._section_name, configuration)
        self._name = configuration_file.get_option(jobstore, "name", missing_value="")
        if self._name in ['mongo', 'redis', 'sqlalchemy']:
            if self._name == 'mongo':
                self._storage = SchedulerMongoJobStoreConfiguration()

            elif self._name == 'redis':
                self._storage = SchedulerRedisJobStoreConfiguration()

            else:
                self._storage = SchedulerSqlAlchemyJobStoreConfiguration()

            self._storage.load_config_section(configuration_file, jobstore, bot_root)

        else:
            YLogger.error(self, "No job store storage defined in scheduler config!")

    def to_yaml(self, data, defaults=True):
        if defaults is True:
            data['name'] = "mongo"
        else:
            data['name'] = self._name

        if data['name'] == 'mongo':
            self.config_to_yaml(data, SchedulerMongoJobStoreConfiguration(), defaults)

        if data['name'] == 'redis':
            self.config_to_yaml(data, SchedulerRedisJobStoreConfiguration(), defaults)

        if data['name'] == 'sqlalchemy':
            self.config_to_yaml(data, SchedulerSqlAlchemyJobStoreConfiguration(), defaults)


class SchedulerMongoJobStoreConfiguration(BaseConfigurationData):

    def __init__(self):
        BaseConfigurationData.__init__(self, name="mongo")
        self._collection = None

    @property
    def collection(self):
        return self._collection

    def load_config_section(self, configuration_file, configuration, bot_root, subs: Substitutions = None):
        del bot_root
        del subs
        mongodb = configuration_file.get_section(self._section_name, configuration)
        if mongodb is not None:
            self._collection = configuration_file.get_option(mongodb, "collection", missing_value=None)

    def to_yaml(self, data, defaults=True):
        if defaults is True:
            data['collection'] = "programy"

        else:
            data['collection'] = self.collection


class SchedulerRedisJobStoreConfiguration(BaseConfigurationData):

    def __init__(self):
        BaseConfigurationData.__init__(self, name="redis")
        self._jobs_key = None
        self._run_times_key = None

    @property
    def jobs_key(self):
        return self._jobs_key

    @property
    def run_times_key(self):
        return self._run_times_key

    def load_config_section(self, configuration_file, configuration, bot_root, subs: Substitutions = None):
        del bot_root
        del subs
        redis = configuration_file.get_section(self._section_name, configuration)
        if redis is not None:
            self._jobs_key = configuration_file.get_option(redis, "jobs_key", missing_value=None)
            self._run_times_key = configuration_file.get_option(redis, "run_times_key", missing_value=None)

    def to_yaml(self, data, defaults=True):
        if defaults is True:
            data['jobs_key'] = "programy.jobs"
            data['run_times_key'] = "programy.run_times"

        else:
            data['jobs_key'] = self.jobs_key
            data['run_times_key'] = self.run_times_key


class SchedulerSqlAlchemyJobStoreConfiguration(BaseConfigurationData):

    def __init__(self):
        BaseConfigurationData.__init__(self, name="sqlalchemy")
        self._url = None

    @property
    def url(self):
        return self._url

    def load_config_section(self, configuration_file, configuration, bot_root, subs: Substitutions = None):
        del bot_root
        del subs
        sqlalchemy = configuration_file.get_section(self._section_name, configuration)
        if sqlalchemy is not None:
            self._url = configuration_file.get_option(sqlalchemy, "url", missing_value=None)

    def to_yaml(self, data, defaults=True):
        if defaults is True:
            data['url'] = 'sqlite:///programy.sqlite'

        else:
            data['url'] = self.url


class SchedulerThreadPoolConfiguration(BaseConfigurationData):

    def __init__(self):
        BaseConfigurationData.__init__(self, name="threadpool")
        self._max_workers = None

    @property
    def max_workers(self):
        return self._max_workers

    def load_config_section(self, configuration_file, configuration, bot_root, subs: Substitutions = None):
        del bot_root
        del subs
        threadpool = configuration_file.get_section(self._section_name, configuration)
        self._max_workers = configuration_file.get_option(threadpool, "max_workers", missing_value=None)

    def to_yaml(self, data, defaults=True):
        if defaults is True:
            data['max_workers'] = 20

        else:
            data['max_workers'] = self.max_workers


class SchedulerProcessPoolConfiguration(BaseConfigurationData):

    def __init__(self):
        BaseConfigurationData.__init__(self, name="processpool")
        self._max_workers = None

    @property
    def max_workers(self):
        return self._max_workers

    def load_config_section(self, configuration_file, configuration, bot_root, subs: Substitutions = None):
        del bot_root
        del subs
        processpool = configuration_file.get_section(self._section_name, configuration)
        self._max_workers = configuration_file.get_option(processpool, "max_workers", missing_value=None)

    def to_yaml(self, data, defaults=True):
        if defaults is True:
            data['max_workers'] = 5

        else:
            data['max_workers'] = self.max_workers


class SchedulerJobDefaultsConfiguration(BaseConfigurationData):

    def __init__(self):
        BaseConfigurationData.__init__(self, name="job_defaults")
        self._coalesce = None
        self._max_instances = None

    @property
    def coalesce(self):
        return self._coalesce

    @property
    def max_instances(self):
        return self._max_instances

    def load_config_section(self, configuration_file, configuration, bot_root, subs: Substitutions = None):
        del bot_root
        del subs
        job_defaults = configuration_file.get_section(self._section_name, configuration)
        self._coalesce = configuration_file.get_option(job_defaults, "coalesce", missing_value=None)
        self._max_instances = configuration_file.get_option(job_defaults, "max_instances", missing_value=None)

    def to_yaml(self, data, defaults=True):
        if defaults is True:
            data['coalesce'] = False
            data['max_instances'] = 3

        else:
            data['coalesce'] = self.coalesce
            data['max_instances'] = self.max_instances


class SchedulerConfiguration(BaseConfigurationData):

    def __init__(self):
        BaseConfigurationData.__init__(self, name="scheduler")
        self._name = None
        self._debug_level = 0
        self._add_listeners = False
        self._remove_all_jobs = False
        self._blocking = False

        self._jobstore = None
        self._threadpool = None
        self._processpool = None
        self._job_defaults = None

    @property
    def name(self):
        return self._name

    @property
    def debug_level(self):
        return self._debug_level

    @property
    def add_listeners(self):
        return self._add_listeners

    @property
    def remove_all_jobs(self):
        return self._remove_all_jobs

    @property
    def blocking(self):
        return self._blocking

    @property
    def jobstore(self):
        return self._jobstore

    @property
    def threadpool(self):
        return self._threadpool

    @property
    def processpool(self):
        return self._processpool

    @property
    def job_defaults(self):
        return self._job_defaults

    def _load_scheduler_settings(self, configuration_file, scheduler, subs: Substitutions = None):
        self._name = configuration_file.get_option(scheduler, "name", missing_value=None, subs=subs)
        self._debug_level = configuration_file.get_int_option(scheduler, "debug_level", missing_value=0, subs=subs)
        self._add_listeners = configuration_file.get_bool_option(scheduler, "add_listeners",
                                                                 missing_value=False, subs=subs)
        self._remove_all_jobs = configuration_file.get_bool_option(scheduler, "remove_all_jobs",
                                                                   missing_value=False, subs=subs)

    def _load_jobstore(self, configuration_file, scheduler, bot_root):
        self._jobstore = SchedulerJobStoreConfiguration()
        self._jobstore.load_config_section(configuration_file, scheduler, bot_root)

    def _load_threadpool(self, configuration_file, scheduler, bot_root):
        self._threadpool = SchedulerThreadPoolConfiguration()
        self._threadpool.load_config_section(configuration_file, scheduler, bot_root)

    def _load_processpool(self, configuration_file, scheduler, bot_root):
        self._processpool = SchedulerProcessPoolConfiguration()
        self._processpool.load_config_section(configuration_file, scheduler, bot_root)

    def _load_job_defaults(self, configuration_file, scheduler, bot_root):
        self._job_defaults = SchedulerJobDefaultsConfiguration()
        self._job_defaults.load_config_section(configuration_file, scheduler, bot_root)

    def _load_scheduler_config(self, configuration_file, scheduler, bot_root, subs: Substitutions = None):

        self._load_scheduler_settings(configuration_file, scheduler, subs)

        if 'jobstore' in scheduler:
            self._load_jobstore(configuration_file, scheduler, bot_root)

        if 'threadpool' in scheduler:
            self._load_threadpool(configuration_file, scheduler, bot_root)

        if 'processpool' in scheduler:
            self._load_processpool(configuration_file, scheduler, bot_root)

        if 'job_defaults' in scheduler:
            self._load_job_defaults(configuration_file, scheduler, bot_root)

    def load_config_section(self, configuration_file, configuration, bot_root, subs: Substitutions = None):
        scheduler = configuration_file.get_section(self._section_name, configuration)
        if scheduler is not None:
            self._load_scheduler_config(configuration_file, scheduler, bot_root, subs)

        else:
            YLogger.warning(self, "'scheduler' section missing from client config, using to defaults")

    def _create_mongo_jobstore_config(self, config):
        config['apscheduler.jobstores.mongo'] = {'type': 'mongodb'}
        if self.jobstore is not None:
            if self.jobstore.storage:
                config['apscheduler.jobstores.mongo']['collection'] = self.jobstore.storage.collection

    def _create_redis_jobstore_config(self, config):
        config['apscheduler.jobstores.redis'] = {'type': 'redis'}
        if self.jobstore is not None:

            if self.jobstore.storage:

                if self.jobstore.storage.jobs_key is not None:
                    config['apscheduler.jobstores.redis']['jobs_key'] = self.jobstore.storage.jobs_key

                if self.jobstore.storage.run_times_key is not None:
                    config['apscheduler.jobstores.redis']['run_times_key'] = self.jobstore.storage.run_times_key

    def _create_sqlalchemy_jobstore_config(self, config):
        config['apscheduler.jobstores.sqlalchemy'] = {'type': 'sqlalchemy'}
        if self.jobstore is not None:

            if self.jobstore.storage:

                if self.jobstore.storage.url is not None:
                    config['apscheduler.jobstores.sqlalchemy']['url'] = self.jobstore.storage.url

    def _create_threadpool_config(self, config):
        config['apscheduler.executors.default'] = {'class': 'apscheduler.executors.pool:ThreadPoolExecutor'}
        if self.threadpool is not None:
            if self.threadpool.max_workers is not None:
                config['apscheduler.executors.default']['max_workers'] = str(self.threadpool.max_workers)

    def _create_processpool_config(self, config):
        config['apscheduler.executors.processpool'] = {'type': 'processpool'}
        if self.processpool is not None:
            if self.processpool.max_workers is not None:
                config['apscheduler.executors.processpool']['max_workers'] = str(self.processpool.max_workers)

    def _create_job_defaults_config(self, config):
        config['apscheduler.job_defaults'] = {}
        if self.job_defaults:
            if self.job_defaults.coalesce is not None:
                config['apscheduler.job_defaults']['coalesce'] = str(self.job_defaults.coalesce).lower()

            if self.job_defaults.max_instances is not None:
                config['apscheduler.job_defaults']['max_instances'] = str(self.job_defaults.max_instances)

    def create_scheduler_config(self):

        config = {}
        if self.jobstore is not None:

            if self.jobstore.name in ['mongo', 'redis', 'sqlalchemy']:

                if self.jobstore.name == 'mongo':
                    self._create_mongo_jobstore_config(config)

                elif self.jobstore.name == 'redis':
                    self._create_redis_jobstore_config(config)

                else:
                    self._create_sqlalchemy_jobstore_config(config)

        if self.threadpool is not None:
            self._create_threadpool_config(config)

        if self.processpool is not None:
            self._create_processpool_config(config)

        if self.job_defaults is not None:
            self._create_job_defaults_config(config)

        if len(config) > 0:
            return config

        return None

    def to_yaml(self, data, defaults=True):
        if defaults is True:
            data['name'] = "scheduler"
            data['debug_level'] = 0
            data['add_listeners'] = False
            data['remove_all_jobs'] = False

        else:
            data['name'] = self.name
            data['debug_level'] = self.debug_level
            data['add_listeners'] = self.add_listeners
            data['remove_all_jobs'] = self.remove_all_jobs

        self.config_to_yaml(data, SchedulerJobStoreConfiguration(), defaults)
        self.config_to_yaml(data, SchedulerThreadPoolConfiguration(), defaults)
        self.config_to_yaml(data, SchedulerProcessPoolConfiguration(), defaults)
        self.config_to_yaml(data, SchedulerJobDefaultsConfiguration(), defaults)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


class Authorisable:

    def __init__(self, identifier):
        self._id = identifier
        self._roles = []
        self._groups = []

    @property
    def roles(self):
        return self._roles

    def add_role(self, role):
        if role not in self._roles:
            self.roles.append(role)

    def has_role(self, role):
        if role in self._roles:
            return True
        for group in self._groups:
            if group.has_role(role) is True:
                return True
        return False

    @property
    def groups(self):
        return self._groups

    def add_groups(self, groups):
        self._groups = groups[:]

    def add_group(self, group):
        if group not in self._groups:
            self._groups.append(group)

    def has_group(self, groupid):
        for group in self._groups:
            if groupid == group.groupid:
                return True
            elif group.has_group(groupid):
                return True
        return False

    def available_roles(self):
        roles = self._roles[:]
        for group in self._groups:
            roles += group.available_roles()
        return roles


class Group(Authorisable):

    def __init__(self, groupid):
        Authorisable.__init__(self, groupid)
        self._users = []

    @property
    def groupid(self):
        return self._id

    @property
    def users(self):
        return self._users

    def add_users(self, users):
        self._users = users[:]

    def add_user(self, user):
        if user not in self._users:
            self._users.append(user)

    def has_user(self, userid):
        for user in self._users:
            if user.userid == userid:
                return True
        return False


class User(Authorisable):

    def __init__(self, userid):
        Authorisable.__init__(self, userid)

    @property
    def userid(self):
        return self._id

    def add_to_group(self, group):
        if group not in self._groups:
            self._groups.append(group)
        group.add_user(self)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger

from programy.config.section import BaseSectionConfigurationData
from programy.utils.substitutions.substitues import Substitutions


class BrainServiceConfiguration(BaseSectionConfigurationData):
    additionals = ['denied_srai']

    def __init__(self, service_name):
        BaseSectionConfigurationData.__init__(self, service_name)
        self._classname = None
        self._method = None
        self._host = None
        self._port = None
        self._url = None

    @property
    def classname(self):
        return self._classname

    @property
    def method(self):
        return self._method

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def url(self):
        return self._url

    def additionals_to_add(self):
        return BrainServiceConfiguration.additionals

    def load_config_section(self, configuration_file, configuration, bot_root, subs: Substitutions = None):
        service = configuration_file.get_section(self.section_name, configuration)
        if service is not None:
            self._classname = configuration_file.get_option(service, "classname", missing_value=None, subs=subs)
            self._method = configuration_file.get_option(service, "method", missing_value=None, subs=subs)
            self._host = configuration_file.get_option(service, "host", missing_value=None, subs=subs)
            self._port = configuration_file.get_int_option(service, "port", missing_value=None, subs=subs)
            self._url = configuration_file.get_option(service, "url", missing_value=None, subs=subs)
            self.load_additional_key_values(configuration_file, service)
        else:
            YLogger.warning(self, "'services' section missing from brain config, using to defaults")

    def to_yaml(self, data, defaults=True):
        data['classname'] = self._classname
        data['method'] = self._method
        data['host'] = self._host
        data['port'] = self._port
        data['url'] = self._url



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import logging
import hashlib
from datetime import datetime
from datetime import timedelta
from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.schedulers.blocking import BlockingScheduler
from apscheduler.events import EVENT_ALL
from apscheduler.events import SchedulerEvent, JobEvent, JobSubmissionEvent, JobExecutionEvent
from programy.utils.logging.ylogger import YLogger


def scheduled(name, userid, clientid, action, text):
    ProgramyScheduler.scheduled_event(name, userid, clientid, action, text)


def scheduler_listener(event):
    ProgramyScheduler.listener_event(event)


class ProgramyScheduler:
    schedulers = {}

    def __init__(self, client, configuration):
        self._client = client
        self._scheduler = None
        self._configuration = configuration

        if self._configuration:
            if self._configuration.debug_level is not None:
                self.set_debug_level(self._configuration.debug_level)

            self._scheduler = self._create_scheduler(self._configuration)

            self.register_schedulder(self, self._configuration.name)

            if self._configuration.add_listeners is True:
                YLogger.debug(None, "Scheduler registering listeners")
                self.register_listeners()

            if self._configuration.remove_all_jobs is True:
                YLogger.debug(None, "Scheduler removing all existing jobs")
                self._scheduler.remove_all_jobs()

            YLogger.debug(None, "Scheduler initiated...")
        else:
            YLogger.debug(None, "Scheduler not initiated...")

    @property
    def name(self):
        return self._configuration.name

    def register_schedulder(self, scheduler, name):
        if name in ProgramyScheduler.schedulers:
            YLogger.error(None, "Scheduler with name [%s] already registered", name)
        else:
            ProgramyScheduler.schedulers[name] = scheduler

    def set_debug_level(self, level):
        logging.getLogger('apscheduler').setLevel(level)

    def register_listeners(self):
        self._scheduler.add_listener(scheduler_listener, EVENT_ALL)

    def _create_scheduler(self, configuration):
        config = configuration.create_scheduler_config()
        if configuration.blocking is True:

            if config is not None:
                YLogger.debug(None, "Creating Blocking Scheduler WITH config")
                return BlockingScheduler(config)

            else:
                YLogger.debug(None, "Creating Blocking Scheduler WITHOUT config")
                return BlockingScheduler()

        else:
            if config is not None:
                YLogger.debug(None, "Creating Background Scheduler WITH config")
                return BackgroundScheduler(config)

            else:
                YLogger.debug(None, "Creating Background Scheduler WITHOUT config")
                return BackgroundScheduler()

    def start(self):
        YLogger.debug(None, "Scheduler starting...")
        self._scheduler.start()

    def stop(self):
        YLogger.debug(None, "Scheduler stopping...")
        self._scheduler.shutdown()

    def stop_job(self, jobid):
        YLogger.debug(None, "Scheduler stopping %s" % jobid)
        self._scheduler.remove_job(jobid)

    def pause(self):
        YLogger.debug(None, "Scheduler pausing all...")
        self._scheduler.pause()

    def pause_job(self, jobid):
        YLogger.debug(None, "Scheduler pausing %s" % jobid)
        self._scheduler.pause_job(jobid)

    def resume(self):
        YLogger.debug(None, "Scheduler resuming all...")
        self._scheduler.resume()

    def resume_job(self, jobid):
        YLogger.debug(None, "Scheduler resuming %s" % jobid)
        self._scheduler.resume_job(jobid)

    def list_jobs(self):
        id_jobs = {}
        jobs = self._scheduler.get_jobs()
        for job in jobs:
            id_jobs[job.id] = job
        return id_jobs

    def create_job_id(self, userid, clientid, action, text):
        jobid = "%s:%s:%s:%s" % (userid, clientid, action, text)
        hashed = hashlib.md5(jobid.encode())
        return hashed.hexdigest()

    def remove_existing_job(self, job_id):
        old_job = self._scheduler.get_job(job_id)
        if old_job is not None:
            self._scheduler.remove_job(old_job)

    #################################################################################################################
    # Interval triggers

    def schedule_every_n_seconds(self, userid, clientid, action, text, seconds):
        YLogger.debug(None, "Scheduler scheduling every %d seconds" % seconds)
        job_id = self.create_job_id(userid, clientid, action, text)
        self.remove_existing_job(job_id)
        self._scheduler.add_job(scheduled, 'interval', [self.name, userid, clientid, action, text], id=job_id,
                                seconds=seconds)

    def schedule_every_n_minutes(self, userid, clientid, action, text, minutes):
        YLogger.debug(None, "Scheduler scheduling every %d minutes" % minutes)
        job_id = self.create_job_id(userid, clientid, action, text)
        self.remove_existing_job(job_id)
        self._scheduler.add_job(scheduled, 'interval', [self.name, userid, clientid, action, text], id=job_id,
                                minutes=minutes)

    def schedule_every_n_hours(self, userid, clientid, action, text, hours):
        YLogger.debug(None, "Scheduler scheduling every %d hours" % hours)
        job_id = self.create_job_id(userid, clientid, action, text)
        self.remove_existing_job(job_id)
        self._scheduler.add_job(scheduled, 'interval', [self.name, userid, clientid, action, text], id=job_id,
                                hours=hours)

    def schedule_every_n_days(self, userid, clientid, action, text, days):
        YLogger.debug(None, "Scheduler scheduling every %d days" % days)
        job_id = self.create_job_id(userid, clientid, action, text)
        self.remove_existing_job(job_id)
        self._scheduler.add_job(scheduled, 'interval', [self.name, userid, clientid, action, text], id=job_id,
                                days=days)

    def schedule_every_n_weeks(self, userid, clientid, action, text, weeks):
        YLogger.debug(None, "Scheduler scheduling every %d weeks" % weeks)
        job_id = self.create_job_id(userid, clientid, action, text)
        self.remove_existing_job(job_id)
        self._scheduler.add_job(scheduled, 'interval', [self.name, userid, clientid, action, text], id=job_id,
                                weeks=weeks)

    def schedule_every_n(self, userid, clientid, action, text, weeks=0, days=0, hours=0, minutes=0, seconds=0):
        YLogger.debug(None, "Scheduler scheduling every n")
        job_id = self.create_job_id(userid, clientid, action, text)
        self.remove_existing_job(job_id)
        self._scheduler.add_job(scheduled, 'interval', [self.name, userid, clientid, action, text], id=job_id,
                                weeks=weeks, days=days,
                                hours=hours, minutes=minutes, seconds=seconds)

    #################################################################################################################
    # Date trigger

    def schedule_in_n_weeks(self, userid, clientid, action, text, weeks):
        YLogger.debug(None, "Scheduler scheduling in %d weeks" % weeks)
        now = datetime.now()
        the_future = now + timedelta(weeks=weeks)
        self.schedule_at_datetime(userid, clientid, action, text, the_future)

    def schedule_in_n_days(self, userid, clientid, action, text, days):
        YLogger.debug(None, "Scheduler scheduling in %d days" % days)
        now = datetime.now()
        the_future = now + timedelta(days=days)
        self.schedule_at_datetime(userid, clientid, action, text, the_future)

    def schedule_in_n_hours(self, userid, clientid, action, text, hours):
        YLogger.debug(None, "Scheduler scheduling in %d hours" % hours)
        now = datetime.now()
        the_future = now + timedelta(hours=hours)
        self.schedule_at_datetime(userid, clientid, action, text, the_future)

    def schedule_in_n_minutes(self, userid, clientid, action, text, minutes):
        YLogger.debug(None, "Scheduler scheduling in %d minutes" % minutes)
        now = datetime.now()
        the_future = now + timedelta(minutes=minutes)
        self.schedule_at_datetime(userid, clientid, action, text, the_future)

    def schedule_in_n_seconds(self, userid, clientid, action, text, seconds):
        YLogger.debug(None, "Scheduler scheduling in %d seconds" % seconds)
        now = datetime.now()
        the_future = now + timedelta(seconds=seconds)
        self.schedule_at_datetime(userid, clientid, action, text, the_future)

    def schedule_at_datetime(self, userid, clientid, action, text, schedule):
        YLogger.debug(None, "Scheduler scheduling in datetime")
        job_id = self.create_job_id(userid, clientid, action, text)
        self.remove_existing_job(job_id)
        self._scheduler.add_job(scheduled, 'date', [self.name, userid, clientid, action, text], id=job_id,
                                run_date=schedule)

    #################################################################################################################
    # Cron triggers

    def schedule_as_cron(self, userid, clientid, action, text, year='*', month='*', day='*', week='*', day_of_week='*',
                         hour='*', minute='*', second='*'):
        YLogger.debug(None, "Scheduler scheduling cron")
        job_id = self.create_job_id(userid, clientid, action, text)
        self.remove_existing_job(job_id)
        self._scheduler.add_job(scheduled, 'cron', [self.name, userid, clientid, action, text], id=job_id, year=year,
                                month=month, day=day,
                                week=week, day_of_week=day_of_week, hour=hour, minute=minute, second=second)

    #################################################################################################################
    # Admin/Debug Functions

    @staticmethod
    def get_event_str(event):
        if isinstance(event, JobExecutionEvent):
            if event.exception:
                return "JobExecutionEvent [%s] [%s] [%s] [%s] [%s] [%s]" % (
                    event.code, event.job_id, event.jobstore, event.scheduled_run_time, event.retval, event.exception)

            else:
                return "JobExecutionEvent [%s] [%s] [%s] [%s] [%s]" % (
                    event.code, event.job_id, event.jobstore, event.scheduled_run_time, event.retval)

        elif isinstance(event, JobSubmissionEvent):
            return "JobSubmissionEvent [%s] [%s] [%s] [%s]" % (
                str(event.code), event.job_id, event.jobstore, event.scheduled_run_times)

        elif isinstance(event, JobEvent):
            return "JobEvent [%s] [%s] [%s] [%s]" % (event.code, event.job_id, event.jobstore, event.alias)

        elif isinstance(event, SchedulerEvent):
            return "SchedulerEvent [%s] [%s]" % (event.code, event.alias)

        return None

    @staticmethod
    def listener_event(event):
        message = ProgramyScheduler.get_event_str(event)
        if message is not None:
            YLogger.debug(None, message)

        else:
            YLogger.error(None, "Unknown APSchedulerEvent! %s", str(event))

    #################################################################################################################
    # Admin/Debug Functions

    @staticmethod
    def scheduled_event(name, userid, clientid, action, text):
        YLogger.debug(None, "Received Scheduled Event [%s] [%s] [%s] [%s] [%s]",
                      name, userid, clientid, action, text)

        if name in ProgramyScheduler.schedulers:
            scheduler = ProgramyScheduler.schedulers[name]
            scheduler.scheduled(userid, clientid, action, text)

        else:
            YLogger.error(None, "Unknown scheduler [%s]", name)

    def scheduled(self, userid, clientid, action, text):
        YLogger.debug(None, "Processing Scheduled Event [%s] [%s] [%s] [%s] [%s]",
                      self.name, userid, clientid, action, text)

        client_context = self._client.create_client_context(userid)
        if action == 'TEXT':
            self._client.render_response(client_context, text)

        elif action == 'SRAI':
            response = client_context.bot.ask_question(client_context, text)
            self._client.render_response(client_context, response)

        else:
            YLogger.error(client_context, "Unknown scheduler command [%s]", action)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os
import os.path
from programy.utils.parsing.linenumxml import LineNumberingParser
import xml.etree.ElementTree as ET  # pylint: disable=wrong-import-order
from programy.storage.stores.file.store.filestore import FileStore
from programy.storage.entities.learnf import LearnfStore
from programy.utils.logging.ylogger import YLogger


class FileLearnfStore(FileStore, LearnfStore):

    def __init__(self, storage_engine):
        FileStore.__init__(self, storage_engine)
        LearnfStore.__init__(self)

    @staticmethod
    def create_learnf_path(client_context, learnf_dir):
        return "%s%s%s.aiml" % (learnf_dir, os.sep, client_context.userid)

    @staticmethod
    def _write_xml_to_learnf_file(learnf_path):
        dirpath = FileStore._get_dir_from_path(learnf_path)
        FileStore._ensure_dir_exists(dirpath)
        with open(learnf_path, "w+", encoding="utf-8") as file:
            file.write('<?xml version="1.0" encoding="UTF-8"?>\n')
            file.write('<aiml>\n')
            file.write('</aiml>\n')
            file.close()

    @staticmethod
    def create_learnf_file_if_missing(learnf_path):
        if os.path.isfile(learnf_path) is False:
            try:
                YLogger.debug(None, "Creating new learnf file [%s]", learnf_path)
                FileLearnfStore._write_xml_to_learnf_file(learnf_path)
                return True

            except Exception as excep:
                YLogger.exception_nostack(None, "Error Writing learnf to %s", excep, learnf_path)
        return False

    @staticmethod
    def write_node_to_learnf_file(client_context, node, learnf_path):

        YLogger.debug(client_context, "Writing learnf to %s", learnf_path)

        try:
            tree = ET.parse(learnf_path)
        except Exception:
            # Assume invalid aiml file, so remove it and start again with a fresh copy
            if os.path.exists(learnf_path):
                os.remove(learnf_path)
            FileLearnfStore.create_learnf_file_if_missing(learnf_path)
            tree = ET.parse(learnf_path)

        root = tree.getroot()

        # Only add a new node if it doesn't already exist
        if FileLearnfStore.node_already_exists(root, node) is False:
            root.append(node)
            tree.write(learnf_path, method="xml")

    @staticmethod
    def node_already_exists(root, node):

        new_pattern = node.find('pattern')
        if new_pattern is None:
            return False

        new_pattern_str = ET.tostring(new_pattern)

        for category in root:
            current_pattern = category.find('pattern')
            if current_pattern is not None:
                current_pattern_str = ET.tostring(current_pattern)
                if current_pattern_str == new_pattern_str:
                    return True

        return False

    def _get_storage_path(self):
        if len(self.storage_engine.configuration.learnf_storage.dirs) > 1:
            YLogger.warning(self, "Learnf Storage has multiple folders specified, using first only")

        return self.storage_engine.configuration.learnf_storage.dirs[0]

    def get_storage(self):
        return self.storage_engine.configuration.learnf_storage

    def create_category_xml_node(self, client_context, category):
        # Add our new element
        child = ET.Element("category")
        child.append(category.pattern)
        child.append(category.topic)
        child.append(category.that)
        xml_category = category.template.xml_tree(client_context)
        child.append(xml_category)
        return child

    def _write_learnf_to_file(self, client_context, category):
        xml_node = self.create_category_xml_node(client_context, category)

        learnf_path = self._get_storage_path()
        self._ensure_dir_exists(learnf_path)

        learnf_fullpath = self.create_learnf_path(client_context, learnf_path)

        self.create_learnf_file_if_missing(learnf_fullpath)

        self.write_node_to_learnf_file(client_context, xml_node, learnf_fullpath)

    def save_learnf(self, client_context, category):
        try:
            self._write_learnf_to_file(client_context, category)
            return True

        except Exception as exc:
            YLogger.exception_nostack(client_context, "Failed to save learnf", exc)

        return False



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.dialog.sentence import Sentence


class Question:

    @staticmethod
    def create_from_text(client_context, text, split=True, srai=False):
        question = Question(srai)
        if split is True:
            question.split_into_sentences(client_context, text)
        else:
            question.sentences.append(Sentence(client_context, text))
        question.recalculate_sentinment_score(client_context)
        return question

    @staticmethod
    def create_from_sentence(sentence: Sentence, srai=False):
        question = Question(srai)
        question.sentences.append(sentence)
        return question

    @staticmethod
    def create_from_question(question, srai=False):
        new_question = Question(srai)
        for each_sentence in question.sentences:
            new_question.sentences.append(each_sentence)
        return new_question

    def __init__(self, srai=False):
        self._srai = srai
        self._sentences = []
        self._properties = {}
        self._current_sentence_no = -1

    def debug_info(self, client_context):
        text = ""
        first = True
        for sentence in self._sentences:
            if first is False:
                text += ", "
            text += sentence.text(client_context)
            text += " = "
            if sentence.response is not None:
                text += sentence.response
            else:
                text += "N/A"
            first = False
        return text

    @property
    def srai(self):
        return self._srai

    @srai.setter
    def srai(self, srai):
        self._srai = srai

    @property
    def sentences(self):
        return self._sentences

    @sentences.setter
    def sentences(self, sentences):
        self._sentences = sentences[:]

    @property
    def properties(self):
        return self._properties

    @properties.setter
    def properties(self, properties):
        self._properties = dict(properties)

    def has_response(self):
        for sentence in self._sentences:
            if sentence.response is not None:
                return True
        return False

    @property
    def current_sentence_no(self):
        return self._current_sentence_no

    @current_sentence_no.setter
    def current_sentence_no(self, current_sentence_no):
        self._current_sentence_no = current_sentence_no

    def set_property(self, name: str, value: str):
        self._properties[name] = value

    def property(self, name: str):
        if name in self._properties:
            return self._properties[name]
        return None

    def sentence(self, num: int):
        if num < len(self._sentences):
            return self._sentences[num]
        raise Exception("Num sentence array violation !")

    def current_sentence(self):
        if not self._sentences:
            raise Exception("Num sentence array violation !")
        return self._sentences[self._current_sentence_no]

    def previous_nth_sentence(self, num):
        if num >= len(self._sentences):
            raise Exception("Num sentence array violation !")
        previous = -1 - num
        return self._sentences[previous]

    def combine_sentences(self, client_context):
        return ". ".join([sentence.text(client_context) for sentence in self._sentences])

    def combine_answers(self):
        return ". ".join([sentence.response for sentence in self.sentences if sentence.response is not None])

    def split_into_sentences(self, client_context, text):
        if client_context.bot.sentence_splitter.is_active():
            if text is not None and text.strip():
                all_sentences = client_context.bot.sentence_splitter.split(text)
                for each_sentence in all_sentences:
                    self._sentences.append(Sentence(client_context, each_sentence))
        else:
            self._sentences.append(Sentence(client_context, text))

    def recalculate_sentinment_score(self, client_context):
        for sentence in self._sentences:
            sentence.calculate_sentinment_score(client_context)

    def calculate_sentinment_score(self):

        positivity = 0.00
        subjectivity = 0.00

        count = 0
        for sentence in self._sentences:
            positivity += sentence.positivity
            subjectivity += sentence.subjectivity
            count += 1

        if count > 0:
            positivity /= count
            subjectivity /= count
        else:
            subjectivity = 0.5

        return positivity, subjectivity

    def to_json(self):

        json_data = {
            "srai": self._srai,
            "sentences": [],
            "current_sentence_no": self._current_sentence_no,
            "properties": self._properties
        }

        for sentence in self._sentences:
            json_data["sentences"].append(sentence.to_json())

        return json_data

    @staticmethod
    def from_json(client_context, json_data):

        question = Question()

        question.srai = json_data["srai"]
        question.current_sentence_no = json_data["current_sentence_no"]
        question.properties = json_data["properties"]

        for sentence in json_data["sentences"]:
            question.sentences.append(Sentence.from_json(client_context, sentence))

        return question



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.parser.pattern.nodes.base import PatternNode
from programy.parser.pattern.equalsmatch import EqualsMatch
from programy.parser.exceptions import ParserException


class PatternISetNode(PatternNode):
    iset_count = 1

    def __init__(self, attribs, text, userid='*'):
        PatternNode.__init__(self, userid)
        self._words = []

        if 'words' in attribs:
            words = attribs['words'].upper()
        elif text:
            words = text.upper()
        else:
            raise ParserException("Invalid iset node, no words specified as attribute or text")

        self._parse_words(words)
        self._iset_name = "iset_%d" % (PatternISetNode.iset_count)
        PatternISetNode.iset_count += 1

    def _parse_words(self, words):
        splits = words.split(",")
        for word in splits:
            self._words.append(word.strip().upper())

    @property
    def words(self):
        return self._words

    @property
    def iset_name(self):
        return self._iset_name

    def is_iset(self):
        return True

    def to_xml(self, client_context, include_user=False):
        string = ""
        if include_user is True:
            string += '<iset userid="%s" words="%s">' % (self.userid, ". ".join(self.words))
        else:
            string += '<iset words="%s">' % ". ".join(self.words)
        string += super(PatternISetNode, self).to_xml(client_context)
        string += "</iset>\n"
        return string

    def to_string(self, verbose=True):
        words_str = ",".join(self._words)
        if verbose is True:
            return "ISET [%s] [%s] words=[%s]" % (self.userid, self._child_count(verbose), words_str)
        return "ISET words=[%s]" % words_str

    def equivalent(self, other):
        if self.userid != other.userid:
            return False

        for word in self.words:
            if word not in other.words:
                return False

        return True

    def equals(self, client_context, words, word_no):
        if self.userid != '*':
            if self.userid != client_context.userid:
                return EqualsMatch(False, word_no)

        word = words.word(word_no)
        if word is not None:
            word = word.upper()
            for set_word in self._words:
                if word == set_word:
                    YLogger.debug(client_context, "Found word [%s] in iset", word)
                    return EqualsMatch(True, word_no, word)

        YLogger.error(client_context, "No word [%s] found in iset", word)
        return EqualsMatch(False, word_no)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.parser.exceptions import ParserException
from programy.parser.pattern.nodes.base import PatternNode
from programy.parser.pattern.nodes.topic import PatternTopicNode
from programy.parser.pattern.nodes.that import PatternThatNode


class PatternWildCardNode(PatternNode):

    def __init__(self, wildcard, userid='*'):
        PatternNode.__init__(self, userid)
        if wildcard not in self.matching_wildcards():
            raise ParserException("%s not in valid wildcards %s" % (wildcard, ", ".join(self.matching_wildcards())))
        self._wildcard = wildcard

    def is_wildcard(self):
        return True

    def can_add(self, new_node):
        if new_node.is_root():
            raise ParserException("Cannot add root node to child node")

    @property
    def wildcard(self):
        return self._wildcard

    def matching_wildcards(self):
        return []

    def invalid_topic_or_that(self, tabs, client_context, word, context, matches_add):
        if word == PatternTopicNode.TOPIC:
            YLogger.debug(client_context, "%sFound a topic at the wrong place....", tabs)
            context.pop_matches(matches_add)
            return True

        if word == PatternThatNode.THAT:
            YLogger.debug(client_context, "%sFound a that at the wrong place....", tabs)
            context.pop_matches(matches_add)
            return True

        return False

    def check_child_is_wildcard(self, tabs, client_context, context, words, word_no, match_type, depth):
        if self._0ormore_hash is not None:
            YLogger.debug(client_context, "%sWildcard # is next node, moving on!", tabs)
            match = self._0ormore_hash.consume(client_context, context, words, word_no + 1, match_type, depth + 1,
                                               parent=True)
            #if match is not None:
            #    return match
            return match

        if self._1ormore_underline is not None:
            YLogger.debug(client_context, "%sWildcard _ is next node, moving on!", tabs)
            match = self._1ormore_underline.consume(client_context, context, words, word_no + 1, match_type, depth + 1)
            if match is not None:
                return match

        if self._0ormore_arrow is not None:
            YLogger.debug(client_context, "%sWildcard ^ is next node, moving on!", tabs)
            match = self._0ormore_arrow.consume(client_context, context, words, word_no + 1, match_type, depth + 1,
                                                parent=True)
            #if match is not None:
            #    return match
            return match

        if self._1ormore_star is not None:
            YLogger.debug(client_context, "%sWildcard * is next node, moving on!", tabs)
            match = self._1ormore_star.consume(client_context, context, words, word_no + 1, match_type, depth + 1)
            if match is not None:
                return match

        return None



#http://norvig.com/spell-correct.html

import re
from collections import Counter
from programy.spelling.base import SpellingChecker
from programy.storage.factory import StorageFactory


class NorvigSpellingChecker(SpellingChecker):

    DefaultLetters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'

    def __init__(self, spelling_config=None):
        SpellingChecker.__init__(self, spelling_config)

        self.words = []
        self.sum_of_words = 0

    def initialise(self, storage_factory):
        self.load_corpus(storage_factory)

    def load_corpus(self, storage_factory):
        if storage_factory.entity_storage_engine_available(StorageFactory.SPELLING_CORPUS) is True:
            spelling_engine = storage_factory.entity_storage_engine(StorageFactory.SPELLING_CORPUS)
            store = spelling_engine.spelling_store()
            store.load_spelling(self)

    def add_corpus(self, all_words):
        self.words = Counter(self._all_words(all_words))
        self.sum_of_words = sum(self.words.values())

    def _all_words(self, text):
        return re.findall(r'\w+', text.upper())

    def _probability(self, word):
        # Probability of `word`.dont handle null well ...
        if self.sum_of_words == 0:
            return 0.0
        return self.words[word] // self.sum_of_words

    def _correction(self, word):
        # Most probable spelling _correction for word.
        return max(self._candidates(word), key=self._probability)

    def _candidates(self, word):
        # Generate possible spelling _corrections for word.
        return self._known([word]) or self._known(self._edits1(word)) or self._known(self._edits2(word)) or [word]

    def _known(self, words):
        # The subset of `words` that appear in the dictionary of WORDS.
        return set(w for w in words if w in self.words)

    def _edits1(self, word):
        # All edits that are one edit away from `word`.
        splits = [(word[:i], word[i:]) for i in range(len(word) + 1)]
        deletes = [L + R[1:] for L, R in splits if R]
        transposes = [L + R[1] + R[0] + R[2:] for L, R in splits if len(R) > 1]
        replaces = [L + c + R[1:] for L, R in splits if R for c in NorvigSpellingChecker.DefaultLetters]
        inserts = [L + c + R for L, R in splits for c in NorvigSpellingChecker.DefaultLetters]
        return set(deletes + transposes + replaces + inserts)

    def _edits2(self, word):
        # All edits that are two edits away from `word`.
        return (e2 for e1 in self._edits1(word) for e2 in self._edits1(e1))

    def correct(self, phrase):
        # split sentence in into word to be check by _correction
        phras = ''
        for wrd in phrase.split():
            phras = phras+self._correction(wrd.upper())+' '
        return phras.strip()



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.utils.classes.loader import ClassLoader
from programy.storage.entities.store import Store
from programy.storage.stores.nosql.mongo.store.mongostore import MongoStore
from programy.storage.entities.processors import ProcessorStore
from programy.storage.stores.nosql.mongo.dao.processor import PreProcessor
from programy.storage.stores.nosql.mongo.dao.processor import PostProcessor
from programy.storage.stores.nosql.mongo.dao.processor import PostQuestionProcessor


class MongoProcessorStore(MongoStore, ProcessorStore):

    def __init__(self, storage_engine):
        MongoStore.__init__(self, storage_engine)
        ProcessorStore.__init__(self)

    def load(self, collector, name=None):
        YLogger.info(self, "Loading %s nodes from Mongo", self.collection_name())

        nodes = self.get_all_nodes()
        for node in nodes:
            try:
                collector.add_processor(ClassLoader.instantiate_class(node['classname'])())

            except Exception as excep:
                YLogger.exception(self, "Failed pre-instantiating Processor [%s]", excep, node['classname'])

    def get_all_nodes(self):
        collection = self.collection()
        return collection.find()

    def _load_processors_from_file(self, filename, verbose):
        count = 0
        success = 0
        with open(filename, "r", encoding="utf-8") as file:
            for line in file:
                if self.process_config_line(line, verbose) is True:
                    success += 1
                count += 1
        return count, success

    def upload_from_file(self, filename, fileformat=Store.TEXT_FORMAT, commit=True, verbose=False):

        YLogger.info(self, "Uploading %s nodes to Mongo from [%s]", self.collection_name(), filename)
        try:
            return self._load_processors_from_file(filename, verbose)

        except Exception as excep:
            YLogger.exception(self, "Failed to load file [%s]", excep, filename)

        return 0, 0

    def process_config_line(self, line, verbose=False):
        line = line.strip()
        if line.startswith('#') is False:
            class_name = line.strip()
            node = self._get_entity(class_name)

            if verbose is True:
                YLogger.debug(self, "Uploading %s node [%s] to Mongo", self.collection_name(), class_name)

            return self.add_document(node)

        return False

    def _get_entity(self, classname):
        raise NotImplementedError()  # pragma: no cover


class MongoPreProcessorStore(MongoProcessorStore):
    PREPROCESSORS = 'preprocessors'

    def __init__(self, storage_engine):
        MongoProcessorStore.__init__(self, storage_engine)

    def collection_name(self):
        return MongoPreProcessorStore.PREPROCESSORS

    def _get_entity(self, classname):
        return PreProcessor(classname)


class MongoPostProcessorStore(MongoProcessorStore):
    POSTPROCESSORS = 'postprocessors'

    def __init__(self, storage_engine):
        MongoProcessorStore.__init__(self, storage_engine)

    def collection_name(self):
        return MongoPostProcessorStore.POSTPROCESSORS

    def _get_entity(self, classname):
        return PostProcessor(classname)


class MongoPostQuestionProcessorStore(MongoProcessorStore):
    POSTQUESTIONPROCESSORS = 'postquestionprocessors'

    def __init__(self, storage_engine):
        MongoProcessorStore.__init__(self, storage_engine)

    def collection_name(self):
        return MongoPostQuestionProcessorStore.POSTQUESTIONPROCESSORS

    def _get_entity(self, classname):
        return PostQuestionProcessor(classname)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Workflow is

1. User is associated with an initial PRIMARY Client, e.g Facebook


2. User decides to link account and ask PY to initiate an action

3. PY asks them to login into primary account and ask for a LINK TOKEN
    As part of the process they are asked to provide a SECRET they know
    PY then provides a LINK TOKEN
    User now has
                PRIMARY ACCOUNT ID
                PRIMARY ACCOUNT NAME
                GIVEN TOKEN
                GENERATED TOKEN

        -> LINK PRIMARY ACCOUNT $USERID $ACCOUNTNAME $GIVENTOKEN

        <- PRIMARY ACCOUNT LINKED $GENERATEDTOKEN
        <-> PRIMARY ACCOUNT LINK FAILED $REASON

    Link has a expirary time, circa 1 hr, after which link expires and now tokens will need to be requested

4. PY now tells them to log into the client they want to link

5. User logs into secondary account and asks to link this to primary account

6. PY Asks for 
                PRIMARY ACCOUNT ID
                PRIMARY ACCOUNT NAME
                GIVEN TOKEN
                GENERATED TOKEN

        -> LINK SECONDARY ACCOUNT $SECONDARY_USERID $SECONDARY_ACCOUNT_NAME $PRIMARY_USERID
                                        $PRIMARY_ACCOUNT_NAME $GIVEN_TOKEN $GENERATED_TOKEN
        
        <- SECONDARY ACCOUNT LINKED
        <- SECONDARY ACCOUNT LINK FAILED $REASON

7. PY Links accounts
"""
import string
import random
import datetime


class BasicAccountLinkerService:
    KEY_CHARS = string.ascii_uppercase + string.digits
    TWENTY_FOUR_HOURS = 24 * 60 * 60
    MAX_RETRIES = 3

    def __init__(self, storage_engine):
        self._storage_engine = storage_engine

    def initialise(self, client):
        pass    # pragma: no cover

    def link_user_to_client(self, userid, clientid):

        assert userid is not None
        assert clientid is not None

        if self._storage_engine.user_store().exists(userid, clientid) is True:
            return True

        if self._storage_engine.user_store().add_user(userid, clientid) is not None:
            return True

        return False

    def linked_accounts(self, userid):

        assert userid is not None

        return self._storage_engine.user_store().get_links(userid)

    def unlink_user_from_client(self, userid, clientid):

        assert userid is not None
        assert clientid is not None

        if self._storage_engine.user_store().remove_user(userid, clientid) is True:
            if self._storage_engine.link_store().remove_link(userid) is True:
                if self._storage_engine.linked_account_store().unlink_accounts(userid) is True:
                    return True

        return False

    def unlink_user_from_all_clients(self, userid):

        assert userid is not None

        if self._storage_engine.user_store().remove_user_from_all_clients(userid) is True:
            if self._storage_engine.link_store().remove_link(userid) is True:
                if self._storage_engine.linked_account_store().unlink_accounts(userid) is True:
                    return True

        return False

    def _generate_key(self, size=8):
        return ''.join(random.choice(BasicAccountLinkerService.KEY_CHARS) for _ in range(size))

    def _generate_expirary(self, lifetime):
        return datetime.datetime.now() + datetime.timedelta(seconds=lifetime)

    def generate_link(self, userid, provided_key, lifetime=TWENTY_FOUR_HOURS):

        assert userid is not None
        assert provided_key is not None

        generated_key = self._generate_key()
        expires = self._generate_expirary(lifetime)

        if self._storage_engine.link_store().create_link(userid, provided_key, generated_key, expires) is not None:
            return generated_key

        return None

    def _has_link_expired(self, link):

        assert link is not None

        now = datetime.datetime.now()
        if now > link.expires:
            return True

        return False

    def _expire_link(self, link):

        assert link is not None

        link.expired = True
        self._storage_engine.link_store().update_link(link)

    def _valid_link_keys(self, link, provided_key, generated_key, max_retries):

        assert link is not None

        if link.generated_key == generated_key:
            if link.provided_key == provided_key:
                if link.retry_count < max_retries:
                    return True

        return False

    def _inc_retry_count(self, link):

        assert link is not None

        link.retry_count += 1
        self._storage_engine.link_store().update_link(link)

    def link_accounts(self, userid, provided_key, generated_key, linked_userid, linked_client):

        assert userid is not None
        assert provided_key is not None
        assert generated_key is not None
        assert linked_userid is not None
        assert linked_client is not None

        link = self._storage_engine.link_store().get_link(userid)
        if link is not None:

            if link.expired is False:

                if self._has_link_expired(link) is True:
                    self._expire_link(link)

                elif self._valid_link_keys(link, provided_key, generated_key,
                                           BasicAccountLinkerService.MAX_RETRIES) is False:
                    self._inc_retry_count(link)

                elif self._storage_engine.user_store().add_user(linked_userid, linked_client) is not None:

                    if self._storage_engine.linked_account_store().link_accounts(userid, linked_userid) is not None:
                        return True

        return False

    def reset_link(self, userid):

        assert userid is not None

        link = self._storage_engine.link_store().get_link(userid)

        if link is not None:
            link.retry_count = 0
            self._storage_engine.link_store().update_link(link)
            return True

        return False

    def primary_account(self, secondary_userid):

        assert secondary_userid is not None

        return self._storage_engine.linked_account_store().primary_account(secondary_userid)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.utils.classes.loader import ClassLoader


class BaseTranslator:

    def initialise(self):
        pass    # pragma: no cover

    def languages(self):
        raise NotImplementedError()  # pragma: no cover

    def supports_language(self, language):
        raise NotImplementedError()  # pragma: no cover

    def language_code(self, code):
        raise NotImplementedError()  # pragma: no cover

    def detect(self, text, default='EN'):
        raise NotImplementedError()  # pragma: no cover

    def translate(self, text, from_lang, to_lang="EN"):
        raise NotImplementedError()  # pragma: no cover

    @staticmethod
    def initiate_translator(translator_config):
        if translator_config.classname is not None:
            try:
                YLogger.info(None, "Loading translator from class [%s]", translator_config.classname)
                translator_config = ClassLoader.instantiate_class(translator_config.classname)
                translator = translator_config()
                translator.initialise()
                return translator
            except Exception as excep:
                YLogger.exception(None, "Failed to initiate translator", excep)
        else:
            YLogger.warning(None, "No configuration setting for translator!")

        return None



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger

from programy.config.base import BaseConfigurationData
from programy.utils.substitutions.substitues import Substitutions


class PingResponderConfig(BaseConfigurationData):

    def __init__(self, name="responder"):
        BaseConfigurationData.__init__(self, name)
        self._name = "Client Ping Responder"
        self._host = None
        self._port = None
        self._url = None
        self._ssl_cert_file = None
        self._ssl_key_file = None
        self._shutdown = None
        self._register = None
        self._unregister = None
        self._debug = False

    @property
    def name(self):
        return self._name

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def ssl_cert_file(self):
        return self._ssl_cert_file

    @property
    def ssl_key_file(self):
        return self._ssl_key_file

    @property
    def url(self):
        return self._url

    @property
    def shutdown(self):
        return self._shutdown

    @property
    def register(self):
        return self._register

    @property
    def unregister(self):
        return self._unregister

    @property
    def debug(self):
        return self._debug

    def load_config_section(self, configuration_file, section, bot_root=None, subs: Substitutions = None):
        del bot_root
        del subs
        responder = configuration_file.get_section(self._section_name, section)
        if responder is not None:
            self._name = configuration_file.get_option(responder, "name")
            self._host = configuration_file.get_option(responder, "host")
            self._port = configuration_file.get_option(responder, "port")
            self._ssl_cert_file = configuration_file.get_option(responder, "ssl_cert_file")
            self._ssl_key_file = configuration_file.get_option(responder, "ssl_key_file")
            self._url = configuration_file.get_option(responder, "url")
            self._shutdown = configuration_file.get_option(responder, "shutdown")
            self._register = configuration_file.get_option(responder, "register")
            self._unregister = configuration_file.get_option(responder, "unregister")
            self._debug = configuration_file.get_option(responder, "debug")

        else:
            YLogger.warning(self, "'responder' section missing from client config, using defaults")

    def to_yaml(self, data, defaults=True):

        assert data is not None

        if defaults is True:
            data['name'] = "Client Ping Responder"
            data['host'] = None
            data['port'] = None
            data['ssl_cert_file'] = None
            data['ssl_key_file'] = None
            data['url'] = None
            data['shutdown'] = None
            data['register'] = None
            data['unregister'] = None
            data['debug'] = False

        else:
            data['name'] = self._name
            data['host'] = self._host
            data['port'] = self._port
            data['ssl_cert_file'] = self._ssl_cert_file
            data['ssl_key_file'] = self._ssl_key_file
            data['url'] = self._url
            data['shutdown'] = self._shutdown
            data['register'] = self._register
            data['unregister'] = self._unregister
            data['debug'] = self._debug



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger

from programy.config.base import BaseConfigurationData
from programy.utils.substitutions.substitues import Substitutions


class BotSentenceJoinerConfiguration(BaseConfigurationData):
    DEFAULT_CLASSNAME = "programy.dialog.joiner.joiner.SentenceJoiner"
    DEFAULT_JOIN_CHARS = ".?!"
    DEFAULT_TERMINATOR = "."

    def __init__(self):
        BaseConfigurationData.__init__(self, name="joiner")
        self._classname = BotSentenceJoinerConfiguration.DEFAULT_CLASSNAME
        self._join_chars = BotSentenceJoinerConfiguration.DEFAULT_JOIN_CHARS
        self._terminator = BotSentenceJoinerConfiguration.DEFAULT_TERMINATOR

    @property
    def classname(self):
        return self._classname

    @property
    def join_chars(self):
        return self._join_chars

    @property
    def terminator(self):
        return self._terminator

    def load_config_section(self, configuration_file, configuration, bot_root, subs: Substitutions = None):
        del bot_root
        joiner = configuration_file.get_section(self._section_name, configuration)
        if joiner is not None:
            self._classname = configuration_file.\
                get_option(joiner, "classname", missing_value=BotSentenceJoinerConfiguration.DEFAULT_CLASSNAME,
                           subs=subs)
            self._join_chars = configuration_file.\
                get_option(joiner, "join_chars", missing_value=BotSentenceJoinerConfiguration.DEFAULT_JOIN_CHARS,
                           subs=subs)
            self._terminator = configuration_file.\
                get_option(joiner, "terminator", missing_value=BotSentenceJoinerConfiguration.DEFAULT_TERMINATOR,
                           subs=subs)
        else:
            YLogger.warning(self, "'joiner' section missing from bot config, using defaults")

    def to_yaml(self, data, defaults=True):
        if defaults is True:
            data['classname'] = BotSentenceJoinerConfiguration.DEFAULT_CLASSNAME
            data['join_chars'] = BotSentenceJoinerConfiguration.DEFAULT_JOIN_CHARS
            data['terminator'] = BotSentenceJoinerConfiguration.DEFAULT_TERMINATOR

        else:
            data['classname'] = self._classname
            data['join_chars'] = self._join_chars
            data['terminator'] = self._terminator
